from unittest.mock import patch

from django.test import TestCase
from django.db.models import Q

from allianceauth.tests.auth_utils import AuthUtils

from app_utils.testdata_factories import UserMainFactory

from charlink.forms import LinkForm
from charlink.app_imports import import_apps
from charlink.models import AppSettings


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestLinkForm(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()

    def test_init_no_perms(self):
        import_apps()
        form = LinkForm(self.user)
        self.assertIn('allianceauth.authentication_default', form.fields)
        self.assertNotIn('allianceauth.corputils_default', form.fields)

    def test_init_with_perms(self):
        import_apps()
        user = AuthUtils.add_permissions_to_user_by_name(['corputils.add_corpstats', "marketmanager.basic_market_browser"], self.user)
        form = LinkForm(user)
        self.assertIn('allianceauth.authentication_default', form.fields)
        self.assertIn('allianceauth.corputils_default', form.fields)
        self.assertIn('marketmanager_corporation', form.fields)
        self.assertIn('marketmanager_character', form.fields)

    def test_init_with_perms_ignore(self):
        import_apps()
        AppSettings.objects.filter(
            Q(app_name='allianceauth.corputils_default') | Q(app_name='marketmanager_corporation')
        ).update(ignored=True)

        user = AuthUtils.add_permissions_to_user_by_name(['corputils.add_corpstats', "marketmanager.basic_market_browser"], self.user)
        form = LinkForm(user)
        self.assertIn('allianceauth.authentication_default', form.fields)
        self.assertNotIn('allianceauth.corputils_default', form.fields)
        self.assertNotIn('marketmanager_corporation', form.fields)
        self.assertIn('marketmanager_character', form.fields)



# Generated by Django 4.2.27 on 2026-01-25 09:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('charlink', '0003_compliancefilter'),
    ]

    operations = [
        migrations.AlterField(
            model_name='appsettings',
            name='default_selection',
            field=models.BooleanField(),
        ),
    ]



from unittest.mock import patch

from django.test import TestCase
from django.urls import reverse

from app_utils.testdata_factories import UserMainFactory

from charlink.app_imports import import_apps
from charlink.models import ComplianceFilter

from securegroups.models import SmartFilter


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestHooks(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.testuser = UserMainFactory()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.html_menu = f"""
            <li class="d-flex flex-wrap m-2 p-2 pt-0 pb-0 mt-0 mb-0 me-0 pe-0">
                <i class="nav-link fas fa-link fa-fw align-self-center me-3 active"></i>
                <a class="nav-link flex-fill align-self-center me-auto active" href="{reverse('charlink:index')}">
                    CharLink
                </a>
            </li>
        """
        cls.html_dashboard = (
            f'<form method="post" action="{reverse("charlink:dashboard_post")}">'
        )

    def test_menu_hook(self):
        import_apps()
        self.client.force_login(self.testuser)

        response = self.client.get(reverse("charlink:index"))
        self.assertContains(response, self.html_menu, html=True)

    def test_dashboard_hook(self):
        import_apps()
        self.client.force_login(self.testuser)

        response = self.client.get(reverse("authentication:dashboard"))
        self.assertContains(response, self.html_dashboard, status_code=200)

    def test_securegroups_hook(self):
        cf = ComplianceFilter.objects.create(name="Test Filter", description="A test filter")

        self.assertEqual(SmartFilter.objects.first().filter_object, cf)



from django import forms

from .app_imports import import_apps


class LinkForm(forms.Form):

    def __init__(self, user, *args, **kwargs):
        super().__init__(*args, **kwargs)
        imported_apps = import_apps()
        self.fields['allianceauth.authentication_default'] = forms.BooleanField(
            required=False,
            initial=True,
            disabled=True,
            label=imported_apps['allianceauth.authentication'].get('default').field_label
        )
        for app, imports in imported_apps.items():
            form_fields = imports.get_form_fields(user)
            if app != 'allianceauth.authentication' and len(form_fields) > 0:
                self.fields.update(form_fields)



from collections import defaultdict

from django.db import models
from django.contrib.auth.models import User
from django.utils.translation import gettext_lazy as gl

from allianceauth.eveonline.models import EveCharacter


class General(models.Model):
    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ('view_corp', 'Can view linked character of members of their corporation.'),
            ('view_alliance', 'Can view linked character of members of their alliance.'),
            ('view_state', 'Can view linked character of members of their auth state.'),
            ('view_admin', 'Can view CharLink Admin page.'),
        )


class AppSettings(models.Model):
    app_name = models.CharField(max_length=255, unique=True)

    ignored = models.BooleanField(default=False)
    default_selection = models.BooleanField()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        from .app_imports import import_apps
        imported_apps = import_apps()
        app, _, unique_id = self.app_name.rpartition('_')
        login_import = imported_apps[app].get(unique_id)
        return str(login_import.field_label) if login_import else self.app_name


# Smart filters

class BaseFilter(models.Model):
    name = models.CharField(max_length=500)  # This is the filters name shown to the admin
    description = models.CharField(max_length=500)  # this is what is shown to the user

    class Meta:
        abstract = True
        default_permissions = ()

    def __str__(self):
        return f"{self.name}: {self.description}"

    def process_filter(self, user: User):  # Single User Pass Fail system
        raise NotImplementedError("Please Create a filter!")

    def audit_filter(self, users):  # Bulk check system that also advises the user with simple messages
        raise NotImplementedError("Please Create an audit function!")


class ComplianceFilter(BaseFilter):
    selected_apps = models.ManyToManyField(AppSettings, related_name='+')

    negate = models.BooleanField(default=False)

    class Meta:
        verbose_name = gl("Smart Filter: Compliance")
        verbose_name_plural = verbose_name
        default_permissions = ()

    def process_filter(self, user: User):
        return self.audit_filter(User.objects.filter(pk=user.pk))[user.pk]['check']

    def audit_filter(self, users):
        from .app_imports import import_apps

        output = defaultdict(lambda: {"message": "", "check": False})
        imported_apps = import_apps()

        queries = []

        for selected_app in self.selected_apps.all():
            app, _, unique_id = selected_app.app_name.rpartition('_')
            try:
                login_import = imported_apps[app].get(unique_id)
            except KeyError:
                return output
            queries.append(login_import.is_character_added_annotation)

        chars = EveCharacter.objects.filter(
            character_ownership__user__in=users
        )

        q = queries[0]
        for query in queries[1:]:
            q = q & query

        chars = (
            chars
            .annotate(has_compliance=q)
            .values(user_pk=models.F('character_ownership__user'))
            .annotate(
                has_compliance=models.Sum('has_compliance', output_field=models.IntegerField())
            )
            .annotate(char_count=models.Count('*'))
        )

        for user_data in chars:
            compliance = user_data['has_compliance'] == user_data['char_count']
            output[user_data['user_pk']] = {
                "message": gl("Meets requirements") if compliance else gl("Does not meet requirements"),
                "check": compliance ^ self.negate
            }

        return output



# Generated by Django 4.2.23 on 2025-08-22 10:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('charlink', '0002_appsettings_alter_general_options'),
    ]

    operations = [
        migrations.CreateModel(
            name='ComplianceFilter',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=500)),
                ('description', models.CharField(max_length=500)),
                ('negate', models.BooleanField(default=False)),
                ('selected_apps', models.ManyToManyField(related_name='+', to='charlink.appsettings')),
            ],
            options={
                'verbose_name': 'Smart Filter: Compliance',
                'verbose_name_plural': 'Smart Filter: Compliance',
                'default_permissions': (),
            },
        ),
    ]



from django.db.models import Exists, OuterRef
from django.contrib.auth.models import Permission

from django.utils import translation
from django.utils.translation import gettext as _
from django.utils.html import format_html
from django.contrib import messages


from structures import __title__, tasks
from structures.models import Owner, Webhook, OwnerCharacter
from structures.app_settings import (
    STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED,
    STRUCTURES_DEFAULT_LANGUAGE,
)

from app_utils.allianceauth import notify_admins
from app_utils.allianceauth import users_with_permission

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.authentication.models import CharacterOwnership

from charlink.app_imports.utils import LoginImport, AppImport


def _add_character(request, token):
    token_char = EveCharacter.objects.get(character_id=token.character_id)
    character_ownership = CharacterOwnership.objects.get(
        user=request.user, character=token_char
    )

    try:
        corporation = EveCorporationInfo.objects.get(
            corporation_id=token_char.corporation_id
        )
    except EveCorporationInfo.DoesNotExist:
        corporation = EveCorporationInfo.objects.create_corporation(
            token_char.corporation_id
        )
    owner, created = Owner.objects.update_or_create(
        corporation=corporation, defaults={"is_active": True}
    )
    owner.add_character(character_ownership)
    if created:
        default_webhooks = Webhook.objects.filter(is_default=True)
        if default_webhooks:
            for webhook in default_webhooks:
                owner.webhooks.add(webhook)
            owner.save()

    if owner.characters.count() == 1:
        tasks.update_all_for_owner.delay(owner_pk=owner.pk, user_pk=request.user.pk)
        messages.info(
            request,
            format_html(
                _(
                    "%(corporation)s has been added with %(character)s "
                    "as sync character. "
                    "We have started fetching structures and notifications "
                    "for this corporation and you will receive a report once "
                    "the process is finished."
                )
                % {"corporation": owner, "character": token_char}
            ),
        )

        if STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED:
            with translation.override(STRUCTURES_DEFAULT_LANGUAGE):
                notify_admins(
                    message=_(
                        "%(corporation)s was added as new "
                        "structure owner by %(user)s."
                    )
                    % {"corporation": owner, "user": token.user.username},
                    title=_("%s: Structure owner added: %s") % (__title__, owner),
                )
    else:
        messages.info(
            request,
            format_html(
                _(
                    "%(character)s has been added to %(corporation)s "
                    "as sync character. "
                    "You now have %(characters_count)d sync character(s) configured."
                )
                % {
                    "corporation": owner,
                    "character": token_char,
                    "characters_count": owner.valid_characters_count(),
                }
            ),
        )
        if STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED:
            with translation.override(STRUCTURES_DEFAULT_LANGUAGE):
                notify_admins(
                    message=_(
                        "%(character)s was added as sync character to "
                        "%(corporation)s by %(user)s.\n"
                        "We now have %(characters_count)d sync character(s) configured."
                    )
                    % {
                        "character": token_char,
                        "corporation": owner,
                        "user": request.user.username,
                        "characters_count": owner.valid_characters_count(),
                    },
                    title=_("%s: Character added to: %s") % (__title__, owner),
                )


def _is_character_added(character: EveCharacter):
    return OwnerCharacter.objects.filter(
        character_ownership__character=character
    ).exists()


def _users_with_perms():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label='structures',
            codename='add_structure_owner'
        )
    )


app_import = AppImport('structures', [
    LoginImport(
        app_label='structures',
        unique_id='default',
        field_label=__title__,
        add_character=_add_character,
        scopes=Owner.get_esi_scopes(),
        check_permissions=lambda user: user.has_perm("structures.add_structure_owner"),
        is_character_added=_is_character_added,
        is_character_added_annotation=Exists(
            OwnerCharacter.objects
            .filter(character_ownership__character_id=OuterRef('pk'))
        ),
        get_users_with_perms=_users_with_perms,
        default_initial_selection=False,
    ),
])



from django.db.models import Exists, OuterRef
from django.contrib.auth.models import Permission
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.models import EveCharacter

from charlink.app_imports.utils import LoginImport, AppImport

from marketmanager.views import CHARACTER_SCOPES, CORPORATION_SCOPES
from app_utils.allianceauth import users_with_permission
from esi.models import Token


def _is_character_added_character_login(character: EveCharacter):
    return (
        Token.objects
        .filter(character_id=character.character_id)
        .require_valid()
        .require_scopes(CHARACTER_SCOPES)
        .exists()
    )


def _is_character_added_corporation_login(character: EveCharacter):
    return (
        Token.objects
        .filter(character_id=character.character_id)
        .require_valid()
        .require_scopes(CORPORATION_SCOPES)
        .exists()
    )


app_import = AppImport('marketmanager', [
    LoginImport(
        app_label='marketmanager',
        unique_id='character',
        field_label=_('Market Manager - Character Login'),
        add_character=lambda request, token: None,
        scopes=CHARACTER_SCOPES,
        check_permissions=lambda user: user.has_perm("marketmanager.basic_market_browser"),
        is_character_added=_is_character_added_character_login,
        is_character_added_annotation=Exists(
            Token.objects
            .filter(character_id=OuterRef('character_id'))
            .require_scopes(CHARACTER_SCOPES)
        ),
        get_users_with_perms=lambda: users_with_permission(Permission.objects.get(content_type__app_label='marketmanager', codename='basic_market_browser'))
    ),
    LoginImport(
        app_label='marketmanager',
        unique_id='corporation',
        field_label=_('Market Manager - Corporation Login'),
        add_character=lambda request, token: None,
        scopes=CORPORATION_SCOPES,
        check_permissions=lambda user: user.has_perm("marketmanager.basic_market_browser"),
        is_character_added=_is_character_added_corporation_login,
        is_character_added_annotation=Exists(
            Token.objects
            .filter(character_id=OuterRef('character_id'))
            .require_scopes(CORPORATION_SCOPES)
        ),
        get_users_with_perms=lambda: users_with_permission(Permission.objects.get(content_type__app_label='marketmanager', codename='basic_market_browser')),
        default_initial_selection=False,
    )
])



from django.db.models import Exists, OuterRef
from django.contrib.auth.models import User
from django.utils.translation import gettext_lazy as _

from allianceauth.authentication.models import CharacterOwnership

from charlink.app_imports.utils import LoginImport, AppImport


app_import = AppImport('allianceauth.authentication', [
    LoginImport(
        app_label='allianceauth.authentication',
        unique_id='default',
        field_label=_('Add Character (default)'),
        add_character=lambda request, token: None,
        scopes=['publicData'],
        check_permissions=lambda user: True,
        is_character_added=lambda character: CharacterOwnership.objects.filter(character=character).exists(),
        is_character_added_annotation=Exists(CharacterOwnership.objects.filter(character_id=OuterRef('pk'))),
        get_users_with_perms=lambda: User.objects.filter(
            Exists(CharacterOwnership.objects.filter(user_id=OuterRef('pk')))
        ),
    )
])



from unittest.mock import patch

from django.test import TestCase, RequestFactory
from django.contrib.messages.storage.fallback import FallbackStorage

from app_utils.testdata_factories import UserMainFactory, EveCorporationInfoFactory, EveCharacterFactory
from app_utils.testing import add_character_to_user

from charlink.imports.structures import _add_character, _is_character_added
from charlink.app_imports import import_apps

from structures.models import Webhook, Owner


class TestAddCharacter(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=["structures.add_structure_owner"])
        cls.character = cls.user.profile.main_character
        cls.token = cls.user.token_set.first()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()

    @patch("structures.tasks.update_all_for_owner.delay")
    def test_ok(self, mock_update_all_for_owner):
        mock_update_all_for_owner.return_value = None

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages
        request.user = self.user

        _add_character(request, self.token)

        self.assertTrue(_is_character_added(self.character))
        mock_update_all_for_owner.assert_called_once()

    @patch('allianceauth.eveonline.managers.EveCorporationManager.create_corporation', wraps=lambda corp_id: EveCorporationInfoFactory(corporation_id=corp_id))
    @patch("structures.tasks.update_all_for_owner.delay")
    def test_missing_corp(self, mock_update_all_for_owner, mock_create_corporation):
        mock_update_all_for_owner.return_value = None

        self.character.corporation.delete()

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages
        request.user = self.user

        _add_character(request, self.token)

        self.assertTrue(_is_character_added(self.character))
        self.assertTrue(mock_update_all_for_owner.called)
        self.assertTrue(mock_create_corporation.called)

    @patch("structures.tasks.update_all_for_owner.delay")
    def test_already_added(self, mock_update_all_for_owner):
        mock_update_all_for_owner.return_value = None

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages
        request.user = self.user

        _add_character(request, self.token)
        _add_character(request, self.token)

        self.assertTrue(_is_character_added(self.character))
        self.assertTrue(mock_update_all_for_owner.called)

    @patch("structures.tasks.update_all_for_owner.delay")
    def test_default_webhooks(self, mock_update_all_for_owner):
        mock_update_all_for_owner.return_value = None

        Webhook.objects.create(is_default=True, name="test", url="https://discordapp.com/api/webhooks/123456/abcdef")

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages
        request.user = self.user

        _add_character(request, self.token)

        self.assertTrue(_is_character_added(self.character))
        mock_update_all_for_owner.assert_called_once()
        self.assertEqual(Owner.objects.first().webhooks.count(), 1)

    @patch("structures.tasks.update_all_for_owner.delay")
    @patch('charlink.imports.structures.STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED', False)
    def test_no_admin_notifications(self, mock_update_all_for_owner):
        mock_update_all_for_owner.return_value = None

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages
        request.user = self.user

        _add_character(request, self.token)

        self.assertTrue(_is_character_added(self.character))
        mock_update_all_for_owner.assert_called_once()

    @patch("structures.tasks.update_all_for_owner.delay")
    def test_second_owner(self, mock_update_all_for_owner):
        mock_update_all_for_owner.return_value = None

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages
        request.user = self.user

        _add_character(request, self.token)

        self.assertTrue(_is_character_added(self.character))
        mock_update_all_for_owner.assert_called_once()

        character2 = EveCharacterFactory(corporation=self.character.corporation)
        add_character_to_user(self.user, character2)

        _add_character(request, self.user.token_set.get(character_id=character2.character_id))

        self.assertTrue(_is_character_added(character2))
        self.assertEqual(Owner.objects.first().valid_characters_count(), 2)
        mock_update_all_for_owner.assert_called_once()

    @patch("structures.tasks.update_all_for_owner.delay")
    @patch('charlink.imports.structures.STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED', False)
    def test_second_owner_no_admin_notifications(self, mock_update_all_for_owner):
        mock_update_all_for_owner.return_value = None

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages
        request.user = self.user

        _add_character(request, self.token)

        self.assertTrue(_is_character_added(self.character))
        mock_update_all_for_owner.assert_called_once()

        character2 = EveCharacterFactory(corporation=self.character.corporation)
        add_character_to_user(self.user, character2)

        _add_character(request, self.user.token_set.get(character_id=character2.character_id))

        self.assertTrue(_is_character_added(character2))
        self.assertEqual(Owner.objects.first().valid_characters_count(), 2)
        mock_update_all_for_owner.assert_called_once()


class TestIsCharacterAdded(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=["structures.add_structure_owner"])
        cls.character = cls.user.profile.main_character
        cls.token = cls.user.token_set.first()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()

    @patch("structures.tasks.update_all_for_owner.delay")
    def test_ok(self, mock_update_all_for_owner):
        mock_update_all_for_owner.return_value = None

        self.assertFalse(_is_character_added(self.character))

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages
        request.user = self.user

        _add_character(request, self.token)

        self.assertTrue(_is_character_added(self.character))


class TestCheckPermissions(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.no_perm_user = UserMainFactory()
        cls.perm_user = UserMainFactory(permissions=["structures.add_structure_owner"])

    def test_ok(self):
        login_import = import_apps()['structures'].get('default')

        self.assertTrue(login_import.check_permissions(self.perm_user))
        self.assertFalse(login_import.check_permissions(self.no_perm_user))


class TestGetUsersWithPerms(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.no_perm_user = UserMainFactory()
        cls.perm_user = UserMainFactory(permissions=["structures.add_structure_owner"])

    def test_ok(self):
        login_import = import_apps()['structures'].get('default')

        users = login_import.get_users_with_perms()
        self.assertEqual(users.count(), 1)
        self.assertEqual(users.first(), self.perm_user)



from django.db import transaction
from django.db.models import Exists, OuterRef
from django.contrib.auth.models import Permission
from django.contrib import messages
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from miningtaxes.models import Character, Stats, AdminCharacter
from miningtaxes import tasks

from allianceauth.eveonline.models import EveCharacter

from charlink.app_imports.utils import LoginImport, AppImport

from app_utils.allianceauth import users_with_permission


def _add_character_basic(request, token):
    eve_character = EveCharacter.objects.get(character_id=token.character_id)
    with transaction.atomic():
        character, _ = Character.objects.update_or_create(eve_character=eve_character)
    tasks.update_character.delay(character_pk=character.pk)
    messages.success(
        request,
        format_html(
            "<strong>{}</strong> has been registered. "
            "Note that it can take a minute until all character data is visible.",
            eve_character,
        ),
    )
    s = Stats.load()
    s.calc_admin_main_json()


def _add_character_admin(request, token):
    eve_character = EveCharacter.objects.get(character_id=token.character_id)
    with transaction.atomic():
        character, _ = AdminCharacter.objects.update_or_create(
            eve_character=eve_character
        )
    tasks.update_admin_character.delay(character_pk=character.pk)
    messages.success(
        request,
        format_html(
            "<strong>{}</strong> has been registered. "
            "Note that it can take a minute until all character data is visible.",
            eve_character,
        ),
    )


def _is_character_added_basic(character: EveCharacter):
    return Character.objects.filter(eve_character=character).exists()


def _is_character_added_admin(character: EveCharacter):
    return AdminCharacter.objects.filter(eve_character=character).exists()


def _users_with_perms_basic():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label='miningtaxes',
            codename='basic_access'
        )
    )


def _users_with_perms_admin():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label='miningtaxes',
            codename='admin_access'
        )
    )


app_import = AppImport('miningtaxes', [
    LoginImport(
        app_label='miningtaxes',
        unique_id='default',
        field_label=_("Mining Taxes"),
        add_character=_add_character_basic,
        scopes=Character.get_esi_scopes(),
        check_permissions=lambda user: user.has_perm("miningtaxes.basic_access"),
        is_character_added=_is_character_added_basic,
        is_character_added_annotation=Exists(
            Character.objects
            .filter(eve_character_id=OuterRef('pk'))
        ),
        get_users_with_perms=_users_with_perms_basic,
    ),
    LoginImport(
        app_label='miningtaxes',
        unique_id='admin',
        field_label=_("Mining Taxes Admin"),
        add_character=_add_character_admin,
        scopes=AdminCharacter.get_esi_scopes(),
        check_permissions=lambda user: user.has_perm("miningtaxes.admin_access"),
        is_character_added=_is_character_added_admin,
        is_character_added_annotation=Exists(
            AdminCharacter.objects
            .filter(eve_character_id=OuterRef('pk'))
        ),
        get_users_with_perms=_users_with_perms_admin,
    ),
])



from django.apps import apps
from django.contrib import admin

from .models import ComplianceFilter


class ComplianceFilterAdmin(admin.ModelAdmin):
    filter_horizontal = ('selected_apps',)


if apps.is_installed('securegroups'):
    admin.site.register(ComplianceFilter, ComplianceFilterAdmin)



from django.db.models import Exists, OuterRef
from django.contrib.auth.models import Permission, User

from corptools.models import CharacterAudit, CorporationAudit
from corptools.tasks import update_character, update_all_corps
from corptools.app_settings import get_character_scopes, CORPTOOLS_APP_NAME
from corptools.views import CORP_REQUIRED_SCOPES

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

from charlink.app_imports.utils import LoginImport, AppImport

from app_utils.allianceauth import users_with_permission

_corp_perms = [
    'corptools.own_corp_manager',
    'corptools.alliance_corp_manager',
    'corptools.state_corp_manager',
    'corptools.global_corp_manager',
    'corptools.holding_corp_structures',
    'corptools.holding_corp_assets',
    'corptools.holding_corp_wallets'
]


def _add_character_charaudit(request, token):
    CharacterAudit.objects.update_or_create(
        character=EveCharacter.objects.get_character_by_id(token.character_id))
    update_character.apply_async(args=[token.character_id], kwargs={
                                 "force_refresh": True}, priority=6)


def _add_character_corp(request, token):
    char = EveCharacter.objects.get_character_by_id(token.character_id)
    corp, created = EveCorporationInfo.objects.get_or_create(corporation_id=char.corporation_id,
                                                             defaults={'member_count': 0,
                                                                       'corporation_ticker': char.corporation_ticker,
                                                                       'corporation_name': char.corporation_name
                                                                       })
    CorporationAudit.objects.update_or_create(corporation=corp)
    update_all_corps.apply_async(priority=6)


def _check_perms_corp(user: User):
    return any(user.has_perm(perm) for perm in _corp_perms)


def _is_character_added_charaudit(character: EveCharacter):
    return CharacterAudit.objects.filter(character=character, active=True).exists()


def _is_character_added_corp(character: EveCharacter):
    return CorporationAudit.objects.filter(corporation__corporation_id=character.corporation_id).exists()


def _users_with_perms_charaudit():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label='corptools',
            codename='view_characteraudit'
        )
    )


def _users_with_perms_corp():
    users = users_with_permission(
        Permission.objects.get(
            content_type__app_label=_corp_perms[0].split('.')[0],
            codename=_corp_perms[0].split('.')[1]
        )
    )
    for perm_str in _corp_perms[1:]:
        users |= users_with_permission(
            Permission.objects.get(
                content_type__app_label=perm_str.split('.')[0],
                codename=perm_str.split('.')[1]
            )
        )

    return users


app_import = AppImport('corptools', [
    LoginImport(
        app_label='corptools',
        unique_id='default',
        field_label=CORPTOOLS_APP_NAME,
        add_character=_add_character_charaudit,
        scopes=get_character_scopes(),
        check_permissions=lambda user: user.has_perm('corptools.view_characteraudit'),
        is_character_added=_is_character_added_charaudit,
        is_character_added_annotation=Exists(
            CharacterAudit.objects
            .filter(character_id=OuterRef('pk'), active=True)
        ),
        get_users_with_perms=_users_with_perms_charaudit,
    ),
    LoginImport(
        app_label='corptools',
        unique_id='structures',
        field_label="Corporation Audit",
        add_character=_add_character_corp,
        scopes=CORP_REQUIRED_SCOPES,
        check_permissions=_check_perms_corp,
        is_character_added=_is_character_added_corp,
        is_character_added_annotation=Exists(
            CorporationAudit.objects
            .filter(corporation__corporation_id=OuterRef('corporation_id'))
        ),
        get_users_with_perms=_users_with_perms_corp,
        default_initial_selection=False,
    )
])



import re
from dataclasses import dataclass
from typing import Callable, List
from collections import defaultdict

from django.db.models import Exists, QuerySet
from django import forms
from django.contrib.auth.models import User
from django.conf import settings
from django.http import HttpRequest

from allianceauth.eveonline.models import EveCharacter
from esi.models import Token

from ..models import AppSettings


@dataclass
class LoginImport:
    """
    The class for implementing a login import for an app.


    There can be multiple imports for an app, like in case of corptools where there is Character Audit and Corporation Audit.

    Args:
        `app_label`: The app label of the app the import is for. It must be in settings.INSTALLED_APPS and must be the same of the `AppImport`.
        `unique_id`: A unique (within the app) identifier for the import. It must be a string and contain only alphanumeric characters and no spaces.
        `field_label`: The label for the field in the form.
        `add_character`: A function that adds the character to the app. It must be a callable that takes a `esi.models.Token` as an argument and performs all the operations needed for adding a character to the application.
        `scopes`: A list of scopes required for the import.
        `check_permissions`: A function that checks if the user has permissions to use the import. It must be a callable that takes a `User` as an argument and returns a boolean.
        `is_character_added`: A function that checks if the character is already added to the app. It must be a callable that takes an EveCharacter as an argument and returns a boolean.
        `is_character_added_annotation`: A django Exists object that checks if the character is already added to the app.
        `get_users_with_perms`: A function that returns a QuerySet of users with permissions to use the import. It must be a callable that takes no arguments and returns a QuerySet of Users.
        `default_initial_selection`: Optional boolean that specifies the initial setting for the default selection of the login option in the form. Default is True. It is important to set this to false for options that are not used by everyday characters, to avoid causing issues with ESI rate limits.
    """
    app_label: str
    unique_id: str
    field_label: str
    add_character: Callable[[HttpRequest, Token], None]
    scopes: List[str]
    check_permissions: Callable[[User], bool]
    is_character_added: Callable[[EveCharacter], bool]
    is_character_added_annotation: Exists
    get_users_with_perms: Callable[[], QuerySet[User]]
    default_initial_selection: bool = True

    def get_query_id(self):
        return f"{self.app_label}_{self.unique_id}"

    def __hash__(self) -> int:
        return hash(self.get_query_id())

    def validate_import(self):
        assert hasattr(self, 'app_label')
        assert hasattr(self, 'unique_id')
        assert hasattr(self, 'field_label')
        assert hasattr(self, 'add_character')
        assert hasattr(self, 'scopes')
        assert hasattr(self, 'check_permissions')
        assert hasattr(self, 'is_character_added')
        assert hasattr(self, 'is_character_added_annotation')
        assert hasattr(self, 'get_users_with_perms')
        assert hasattr(self, 'default_initial_selection')
        assert isinstance(self.app_label, str)
        assert isinstance(self.unique_id, str)
        assert re.match(r'^[a-zA-Z0-9]+$', self.unique_id) is not None
        assert isinstance(str(self.field_label), str)
        assert callable(self.add_character)
        assert isinstance(self.scopes, list)
        assert callable(self.check_permissions)
        assert callable(self.is_character_added)
        assert isinstance(self.is_character_added_annotation, Exists)
        assert callable(self.get_users_with_perms)
        assert isinstance(self.default_initial_selection, bool)

    @property
    def is_ignored(self) -> bool:
        app_settings = AppSettings.objects.get(app_name=self.get_query_id())
        return app_settings.ignored

    @property
    def default_selection(self) -> bool:
        app_settings = AppSettings.objects.get(app_name=self.get_query_id())
        return app_settings.default_selection


@dataclass
class AppImport:
    """
    Class wrapper for LoginImports.

    Args:
        `app_label`: The app label of the app the imports are for. It must be in settings.INSTALLED_APPS.
        `imports`: The imports for the app. Must be a list of LoginImport objects.
    """

    app_label: str
    imports: List[LoginImport]

    def get_form_fields(self, user):
        return {
            import_.get_query_id(): forms.BooleanField(
                required=False,
                initial=import_.default_selection,
                label=import_.field_label
            )
            for import_ in self.imports
            if import_.check_permissions(user) and not import_.is_ignored
        }

    def get_imports_with_perms(self, user: User):
        return AppImport(
            self.app_label,
            [
                import_
                for import_ in self.imports
                if import_.check_permissions(user) and not import_.is_ignored
            ]
        )

    def has_any_perms(self, user: User):
        return any(import_.check_permissions(user) for import_ in self.imports if not import_.is_ignored)

    def get(self, unique_id: str) -> LoginImport:
        for import_ in self.imports:
            if import_.unique_id == unique_id:
                return import_

        raise KeyError(f"Import with unique_id {unique_id} not found")

    def validate_import(self):
        assert hasattr(self, 'app_label')
        assert hasattr(self, 'imports')
        assert isinstance(self.app_label, str)
        assert isinstance(self.imports, list)
        assert len(self.imports) > 0
        assert self.app_label in settings.INSTALLED_APPS

        ids = defaultdict(int)

        for import_ in self.imports:
            import_.validate_import()
            ids[import_.unique_id] += 1
            assert import_.app_label == self.app_label

        for count in ids.values():
            assert count == 1



from unittest.mock import patch

from django.test import TestCase, RequestFactory
from django.contrib.messages.storage.fallback import FallbackStorage

from allianceauth.eveonline.models import EveCharacter

from app_utils.testdata_factories import UserMainFactory
from app_utils.testing import create_authgroup

from charlink.imports.memberaudit import _add_character, _is_character_added
from charlink.app_imports import import_apps

from memberaudit.app_settings import MEMBERAUDIT_TASKS_NORMAL_PRIORITY
from memberaudit.models import ComplianceGroupDesignation


class TestAddCharacter(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=["memberaudit.basic_access"])
        cls.character = cls.user.profile.main_character

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()

    @patch('memberaudit.tasks.update_character.apply_async')
    def test_ok(self, mock_update_character):
        mock_update_character.return_value = None

        token = self.user.token_set.first()

        request = self.factory.get('/charlink/login/')
        request.user = self.user
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages

        _add_character(request, token)

        mock_update_character.assert_called_once()
        self.assertTrue(_is_character_added(self.character))

    @patch('memberaudit.tasks.update_compliance_groups_for_user.apply_async')
    @patch('memberaudit.tasks.update_character.apply_async')
    def test_ok_compliance(self, mock_update_character, mock_update_compliance):
        mock_update_character.return_value = None
        mock_update_compliance.return_value = None

        token = self.user.token_set.first()
        request = self.factory.get('/charlink/login/')
        request.user = self.user
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages

        group = create_authgroup()
        ComplianceGroupDesignation.objects.create(group=group)

        _add_character(request, token)

        mock_update_character.assert_called_once()
        mock_update_compliance.assert_called_once_with(
            args=[self.user.pk],
            priority=MEMBERAUDIT_TASKS_NORMAL_PRIORITY
        )
        self.assertTrue(_is_character_added(self.character))


class TestIsCharacterAdded(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=["memberaudit.basic_access"])
        cls.character = cls.user.profile.main_character

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()

    @patch('memberaudit.tasks.update_character.apply_async')
    def test_ok(self, mock_update_character):
        mock_update_character.return_value = None
        login_import = import_apps()['memberaudit'].get('default')

        request = self.factory.get('/charlink/login/')
        request.user = self.user
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages

        self.assertFalse(_is_character_added(self.character))
        self.assertFalse(
            EveCharacter.objects
            .annotate(added=login_import.is_character_added_annotation)
            .get(pk=self.character.pk)
            .added
        )

        _add_character(request, self.user.token_set.first())
        self.assertTrue(_is_character_added(self.character))
        self.assertTrue(
            EveCharacter.objects
            .annotate(added=login_import.is_character_added_annotation)
            .get(pk=self.character.pk)
            .added
        )

        self.character.memberaudit_character.is_disabled = True
        self.character.memberaudit_character.save()

        self.assertFalse(_is_character_added(self.character))
        self.assertFalse(
            EveCharacter.objects
            .annotate(added=login_import.is_character_added_annotation)
            .get(pk=self.character.pk)
            .added
        )


class TestCheckPermissions(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.no_perm_user = UserMainFactory()
        cls.perm_user = UserMainFactory(permissions=["memberaudit.basic_access"])

    def test_ok(self):
        login_import = import_apps()['memberaudit'].get('default')

        self.assertTrue(login_import.check_permissions(self.perm_user))
        self.assertFalse(login_import.check_permissions(self.no_perm_user))


class TestGetUsersWithPerms(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.no_perm_user = UserMainFactory()
        cls.perm_user = UserMainFactory(permissions=["memberaudit.basic_access"])

    def test_ok(self):
        login_import = import_apps()['memberaudit'].get('default')

        users = login_import.get_users_with_perms()
        self.assertEqual(users.count(), 1)
        self.assertEqual(users.first(), self.perm_user)



from django.contrib.auth.models import Permission
from django.contrib import messages
from django.db.models import Exists, OuterRef

from allianceauth.eveonline.models import EveAllianceInfo, EveCharacter, EveCorporationInfo

from aa_contacts.models import AllianceToken, CorporationToken
from aa_contacts.tasks import update_alliance_contacts, update_corporation_contacts

from esi.models import Token

from app_utils.allianceauth import users_with_permission

from ..app_imports.utils import LoginImport, AppImport

ALLIANCE_SCOPES = ['esi-alliances.read_contacts.v1']
CORPORATION_SCOPES = ['esi-corporations.read_contacts.v1']


def _alliance_login(request, token: Token):
    char = EveCharacter.objects.get(character_id=token.character_id)

    if char.alliance_id is None:
        messages.error(request, 'Character is not in an alliance')
        assert False

    try:
        alliance = char.alliance
    except EveAllianceInfo.DoesNotExist:
        alliance = EveAllianceInfo.objects.create_alliance(char.alliance_id)

    if AllianceToken.objects.filter(alliance=alliance).exists():
        messages.error(request, f'{alliance} already has a token')
        assert False

    AllianceToken.objects.create(alliance=alliance, token=token)
    update_alliance_contacts.delay(alliance.alliance_id)


def _corporation_login(request, token: Token):
    char = EveCharacter.objects.get(character_id=token.character_id)

    try:
        corporation = char.corporation
    except EveCorporationInfo.DoesNotExist:
        corporation = EveCorporationInfo.objects.create_corporation(char.corporation_id)

    if CorporationToken.objects.filter(corporation=corporation).exists():
        messages.error(request, f'{corporation} already has a token')
        assert False

    CorporationToken.objects.create(corporation=corporation, token=token)
    update_corporation_contacts.delay(corporation.corporation_id)


def _alliance_users_with_perms():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label='aa_contacts',
            codename='manage_alliance_contacts'
        )
    )


def _corporation_users_with_perms():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label='aa_contacts',
            codename='manage_corporation_contacts'
        )
    )


def _alliance_check_perms(user):
    return user.has_perm('aa_contacts.manage_alliance_contacts')


def _corporation_check_perms(user):
    return user.has_perm('aa_contacts.manage_corporation_contacts')


def _alliance_is_character_added(char: EveCharacter):
    return AllianceToken.objects.filter(token__character_id=char.character_id).exists()


def _corporation_is_character_added(char: EveCharacter):
    return CorporationToken.objects.filter(token__character_id=char.character_id).exists()


app_import = AppImport(
    "aa_contacts",
    [
        LoginImport(
            app_label="aa_contacts",
            unique_id="alliance",
            field_label="Alliance Contacts",
            add_character=_alliance_login,
            scopes=ALLIANCE_SCOPES,
            check_permissions=_alliance_check_perms,
            is_character_added=_alliance_is_character_added,
            is_character_added_annotation=Exists(
                AllianceToken.objects.filter(
                    token__character_id=OuterRef('character_id'),
                )
            ),
            get_users_with_perms=_alliance_users_with_perms,
        ),
        LoginImport(
            app_label="aa_contacts",
            unique_id="corporation",
            field_label="Corporation Contacts",
            add_character=_corporation_login,
            scopes=CORPORATION_SCOPES,
            check_permissions=_corporation_check_perms,
            is_character_added=_corporation_is_character_added,
            is_character_added_annotation=Exists(
                CorporationToken.objects.filter(
                    token__character_id=OuterRef('character_id'),
                )
            ),
            get_users_with_perms=_corporation_users_with_perms,
        ),
    ]
)



from importlib import import_module

from django.conf import settings
from django.utils.translation import gettext_lazy as _

from allianceauth.services.hooks import get_extension_logger
from allianceauth.hooks import get_hooks

from .utils import AppImport
from ..models import AppSettings

logger = get_extension_logger(__name__)

_supported_apps: dict[str, AppImport] = {}
_duplicated_apps = set()
_failed_to_import = {}
_no_import = []

_imported = False


def import_apps():
    global _imported
    if not _imported:
        _supported_apps.clear()
        _duplicated_apps.clear()
        _failed_to_import.clear()
        _no_import.clear()

        # hooks
        charlink_hooks = get_hooks('charlink')

        for hook_f in charlink_hooks:
            hook_mod = hook_f()
            try:
                assert isinstance(hook_mod, str)
                app_import: AppImport = import_module(hook_mod).app_import
                assert type(app_import) == AppImport
                app_import.validate_import()
            except AssertionError:
                logger.debug(f"Loading of {hook_mod} link via hook: failed to validate")
                _failed_to_import[hook_mod] = _("Hook import: failed to validate")
            except ModuleNotFoundError:
                logger.debug(f"Loading of {hook_mod} link via hook: failed to import")
                _failed_to_import[hook_mod] = _("Hook import: import not found")
            except:
                logger.debug(f"Loading of {hook_mod} link via hook: failed")
                _failed_to_import[hook_mod] = _("Hook import: generic error")
            else:
                if app_import.app_label in _supported_apps:
                    _supported_apps.pop(app_import.app_label)
                    _duplicated_apps.add(app_import.app_label)

                if app_import.app_label in _duplicated_apps:
                    logger.debug(f"Loading of {hook_mod} link via hook: failed, duplicate {app_import.app_label}")
                else:
                    _supported_apps[app_import.app_label] = app_import
                    logger.debug(f"Loading of {hook_mod} link via hook: success")

        # defaults
        for app in settings.INSTALLED_APPS:
            if app != 'allianceauth' and app not in _supported_apps:
                try:
                    module = import_module(f'charlink.imports.{app}')
                except ModuleNotFoundError:
                    logger.debug(f"Loading of {app} link: not found")
                    _no_import.append(app)
                else:
                    _supported_apps[app] = module.app_import
                    logger.debug(f"Loading of {app} link: success")

        for app, app_import in _supported_apps.items():
            for login_import in app_import.imports:
                AppSettings.objects.get_or_create(
                    app_name=login_import.get_query_id(),
                    defaults={
                        'default_selection': login_import.default_initial_selection,
                    }
                )

        _imported = True

    return _supported_apps


def get_duplicated_apps():
    import_apps()
    return _duplicated_apps


def get_failed_to_import():
    import_apps()
    return _failed_to_import


def get_no_import():
    import_apps()
    return _no_import



from allianceauth import hooks
from allianceauth.services.hooks import UrlHook, MenuItemHook

from . import urls
from .views import dashboard_login
from .models import ComplianceFilter


class CharlinkMenuItemHook(MenuItemHook):
    def __init__(self):
        super().__init__("CharLink", "fas fa-link", "charlink:index", navactive=['charlink:'])


class LoginDashboardHook(hooks.DashboardItemHook):
    def __init__(self):
        super().__init__(
            dashboard_login,
            6
        )


@hooks.register('menu_item_hook')
def register_menu():
    return CharlinkMenuItemHook()


@hooks.register('url_hook')
def register_urls():
    return UrlHook(urls, 'charlink', 'charlink/')


@hooks.register('dashboard_hook')
def register_login_hook():
    return LoginDashboardHook()


@hooks.register('secure_group_filters')
def securegroups_filters():
    return [ComplianceFilter]



import re

from django.shortcuts import render, redirect, get_object_or_404
from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.auth.models import User
from django.contrib import messages
from django.db.models import Q
from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.template.loader import render_to_string
from django.utils.translation import gettext as _
from django.utils.html import format_html

from allianceauth.services.hooks import get_extension_logger
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.authentication.decorators import permissions_required

from .forms import LinkForm
from .app_imports import import_apps, get_duplicated_apps, get_failed_to_import, get_no_import
from .app_imports.utils import LoginImport
from .decorators import charlink
from .utils import get_user_available_apps, get_user_linked_chars, get_visible_corps, chars_annotate_linked_apps
from .models import AppSettings

logger = get_extension_logger(__name__)


def get_navbar_elements(user: User):
    is_auditor = user.has_perm('charlink.view_state') or user.has_perm('charlink.view_corp') or user.has_perm('charlink.view_alliance')

    return {
        'is_auditor': is_auditor,
        'available_apps': get_user_available_apps(user) if is_auditor else [],
        'available': get_visible_corps(user) if is_auditor else [],
    }


def dashboard_login(request):
    form = LinkForm(request.user, prefix='charlink')
    context = {
        'form': form,
    }
    return render_to_string('charlink/dashboard_login.html', context=context, request=request)


@login_required
def dashboard_post(request):
    if request.method != 'POST':
        messages.error(request, _('Invalid request'))
        return redirect('charlink:index')

    imported_apps = import_apps()

    form = LinkForm(request.user, request.POST, prefix='charlink')
    if not form.is_valid():
        messages.error(request, _('Invalid form data'))
        return redirect('authentication:dashboard')

    scopes = set()
    selected_apps = []

    form_field_pattern = re.compile(r'^(?P<app>[\w\d\.]+)_(?P<unique_id>[a-zA-Z0-9]+)$')

    for import_code, to_import in form.cleaned_data.items():
        if to_import:
            match = form_field_pattern.match(import_code)

            app = match.group('app')
            unique_id = match.group('unique_id')

            app_import = imported_apps[app].get(unique_id)
            scopes.update(app_import.scopes)
            selected_apps.append((app, unique_id))

    request.session['charlink'] = {
        'scopes': list(scopes),
        'imports': selected_apps,
    }

    return redirect('charlink:login')


@login_required
def index(request):
    imported_apps = import_apps()

    if request.method == 'POST':
        form = LinkForm(request.user, request.POST)
        if form.is_valid():
            scopes = set()
            selected_apps = []

            form_field_pattern = re.compile(r'^(?P<app>[\w\d\.]+)_(?P<unique_id>[a-zA-Z0-9]+)$')

            for import_code, to_import in form.cleaned_data.items():
                if to_import:
                    match = form_field_pattern.match(import_code)

                    app = match.group('app')
                    unique_id = match.group('unique_id')

                    app_import = imported_apps[app].get(unique_id)
                    scopes.update(app_import.scopes)
                    selected_apps.append((app, unique_id))

            request.session['charlink'] = {
                'scopes': list(scopes),
                'imports': selected_apps,
            }

            return redirect('charlink:login')

    else:
        form = LinkForm(request.user)

    context = {
        'form': form,
        'characters_added': get_user_linked_chars(request.user),
        **get_navbar_elements(request.user),
    }

    return render(request, 'charlink/charlink.html', context=context)


@login_required
@charlink
def login_view(request, token):
    imported_apps = import_apps()

    charlink_data = request.session.pop('charlink')

    for app, unique_id in charlink_data['imports']:
        import_ = imported_apps[app].get(unique_id)
        if app != 'allianceauth.authentication' and not import_.is_ignored and import_.check_permissions(request.user):
            try:
                import_.add_character(request, token)
            except Exception as e:
                logger.exception(e)
                messages.error(request, _("Failed to add character to %(field_label)s") % {'field_label': import_.field_label})
            else:
                messages.success(request, _("Character successfully added to %(field_label)s") % {'field_label': import_.field_label})

    return redirect('charlink:index')


@login_required
@permissions_required([
    'charlink.view_corp',
    'charlink.view_alliance',
    'charlink.view_state',
])
def audit(request, corp_id: int):
    corp = get_object_or_404(EveCorporationInfo, corporation_id=corp_id)
    corps = get_visible_corps(request.user)

    if not corps.filter(corporation_id=corp_id).exists():
        raise PermissionDenied(_('You do not have permission to view the selected corporation statistics.'))

    context = {
        'selected': corp,
        **get_navbar_elements(request.user),
    }

    return render(request, 'charlink/audit.html', context=context)


@login_required
@permissions_required([
    'charlink.view_corp',
    'charlink.view_alliance',
    'charlink.view_state',
])
def search(request):
    search_string = request.GET.get('search_string', None)
    if not search_string:
        return redirect('charlink:index')

    corps = get_visible_corps(request.user)

    characters = (
        EveCharacter.objects
        .filter(
            character_name__icontains=search_string,
            corporation_id__in=corps.values('corporation_id'),
        )
        .order_by('character_name')
        .select_related('character_ownership__user__profile__main_character')
    )

    context = {
        'search_string': search_string,
        'characters': characters,
        **get_navbar_elements(request.user),
    }

    return render(request, 'charlink/search.html', context=context)


@login_required
@permissions_required([
    'charlink.view_corp',
    'charlink.view_alliance',
    'charlink.view_state',
])
def audit_user(request, user_id):
    user = get_object_or_404(User, pk=user_id)

    corps = get_visible_corps(request.user)

    if (
        not request.user.is_superuser
        and
        user != request.user
        and
        not corps
        .filter(
            corporation_id=user.profile.main_character.corporation_id
        )
        .exists()
    ):
        raise PermissionDenied(_('You do not have permission to view the selected user statistics.'))

    context = {
        'characters_added': get_user_linked_chars(user),
        **get_navbar_elements(request.user),
    }

    return render(request, 'charlink/user_audit.html', context=context)


@login_required
@permissions_required([
    'charlink.view_corp',
    'charlink.view_alliance',
    'charlink.view_state',
])
def audit_app(request, app):
    imported_apps = import_apps()

    if app not in imported_apps:
        raise Http404()

    app_imports = imported_apps[app]

    if not app_imports.has_any_perms(request.user):
        raise PermissionDenied(_('You do not have permission to view the selected application statistics.'))

    app_imports = app_imports.get_imports_with_perms(request.user)

    corp_ids = get_visible_corps(request.user).values('corporation_id')

    logins = {}

    for import_ in app_imports.imports:
        visible_characters = EveCharacter.objects.filter(
            (
                Q(corporation_id__in=corp_ids) |
                Q(character_ownership__user__profile__main_character__corporation_id__in=corp_ids)
            ) &
            Q(character_ownership__user__in=import_.get_users_with_perms()),
        ).select_related('character_ownership__user__profile__main_character')

        visible_characters = chars_annotate_linked_apps(
            visible_characters,
            [import_]
        ).order_by(import_.get_query_id(), 'character_name')

        logins[import_] = visible_characters

    context = {
        'logins': logins,
        'app': app,
        **get_navbar_elements(request.user),
    }

    return render(request, 'charlink/app_audit.html', context=context)


@login_required
@permission_required('charlink.view_admin')
def admin_imported_apps(request):
    context = {
        'imported_apps': import_apps(),
        'duplicated_apps': get_duplicated_apps(),
        'failed_to_import': get_failed_to_import(),
        'no_import': get_no_import(),
        **get_navbar_elements(request.user),
    }

    return render(request, 'charlink/admin_imported_apps.html', context=context)


@login_required
@permission_required('charlink.view_admin')
def toggle_app_visible(request, app_name):
    if app_name == 'allianceauth.authentication_default':
        messages.error(
            request,
            _("The visibility for the built-in AllianceAuth login cannot be changed.")
        )
    else:
        app_settings = get_object_or_404(AppSettings, app_name=app_name)
        app_settings.ignored = not app_settings.ignored
        app_settings.save()

        messages.success(
            request,
            _('App "%(app_name)s" is now %(ignored)s') % {
                'app_name': app_settings,
                'ignored': _('hidden') if app_settings.ignored else _('visible')
            }
        )

    return redirect('charlink:admin_imported_apps')


@login_required
@permission_required('charlink.view_admin')
def toggle_app_default_selection(request, app_name):
    if app_name == 'allianceauth.authentication_default':
        messages.error(
            request,
            _("The default selection for the built-in AllianceAuth login cannot be changed.")
        )
    else:
        app_settings = get_object_or_404(AppSettings, app_name=app_name)
        app_settings.default_selection = not app_settings.default_selection
        app_settings.save()

        imported_apps = import_apps()
        app, _sep, unique_id = app_name.rpartition('_')
        login_import: LoginImport = imported_apps[app].get(unique_id)

        messages.success(
            request,
            _('App "%(app_name)s" is now %(selected)s by default') % {
                'app_name': app_settings,
                'selected': _('selected') if app_settings.default_selection else _('not selected')
            }
        )

        if not login_import.default_initial_selection and app_settings.default_selection:
            messages.warning(
                request,
                format_html(
                    "<h1>" +
                    _("WARNING") +
                    "</h1><br>" +
                    _("The developer set this to not selected in order to avoid ESI bans. Ensure you know what you are doing.")
                )
            )

    return redirect('charlink:admin_imported_apps')



from django.db import transaction
from django.db.models import Exists, OuterRef
from django.contrib.auth.models import Permission
from django.contrib import messages
from django.utils.html import format_html

from esi.models import Token

from memberaudit.models import Character, ComplianceGroupDesignation
from memberaudit.app_settings import MEMBERAUDIT_APP_NAME, MEMBERAUDIT_TASKS_NORMAL_PRIORITY
from memberaudit import tasks

from allianceauth.eveonline.models import EveCharacter

from charlink.app_imports.utils import LoginImport, AppImport

from app_utils.allianceauth import users_with_permission


def _add_character(request, token: Token):
    eve_character = EveCharacter.objects.get(character_id=token.character_id)
    with transaction.atomic():
        character, created = Character.objects.update_or_create(
            eve_character=eve_character, defaults={"is_disabled": False}
        )
    tasks.update_character.apply_async(
        kwargs={
            "character_pk": character.pk,
            "force_update": True,
            "ignore_stale": True,
        },
        priority=MEMBERAUDIT_TASKS_NORMAL_PRIORITY,
    )
    messages.success(
        request,
        format_html(
            "<strong>{}</strong> {}",
            eve_character,
            (
                "has been registered. "
                "Note that it can take a minute until all character data is visible."
            ),
        ),
    )
    if ComplianceGroupDesignation.objects.exists():
        tasks.update_compliance_groups_for_user.apply_async(
            args=[token.user.pk], priority=MEMBERAUDIT_TASKS_NORMAL_PRIORITY
        )


def _is_character_added(character: EveCharacter):
    return Character.objects.filter(eve_character=character, is_disabled=False).exists()


def _users_with_perms():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label='memberaudit',
            codename='basic_access'
        )
    )


app_import = AppImport('memberaudit', [
    LoginImport(
        app_label='memberaudit',
        unique_id='default',
        field_label=MEMBERAUDIT_APP_NAME,
        add_character=_add_character,
        scopes=Character.get_esi_scopes(),
        check_permissions=lambda user: user.has_perm("memberaudit.basic_access"),
        is_character_added=_is_character_added,
        is_character_added_annotation=Exists(
            Character.objects
            .filter(eve_character_id=OuterRef('pk'), is_disabled=False)
        ),
        get_users_with_perms=_users_with_perms,
    ),
])



from django.db.models import Exists, OuterRef
from django.contrib.auth.models import Permission

from moonstuff.providers import ESI_CHARACTER_SCOPES
from moonstuff.models import TrackingCharacter
from moonstuff.tasks import import_extraction_data

from allianceauth.eveonline.models import EveCharacter

from charlink.app_imports.utils import LoginImport, AppImport

from app_utils.allianceauth import users_with_permission


def _add_character(request, token):
    eve_char = EveCharacter.objects.get(character_id=token.character_id)

    if not TrackingCharacter.objects.filter(character=eve_char).exists():
        char = TrackingCharacter(character=eve_char)
        char.save()

        # Schedule an import task to pull data from the new Tracking Character.
        import_extraction_data.delay()
    else:
        assert False


def _is_character_added(character: EveCharacter):
    return TrackingCharacter.objects.filter(character=character).exists()


def _users_with_perms():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label='moonstuff',
            codename='add_trackingcharacter'
        )
    )


app_import = AppImport('moonstuff', [
    LoginImport(
        app_label='moonstuff',
        unique_id='default',
        field_label='Moon Tools',
        add_character=_add_character,
        scopes=ESI_CHARACTER_SCOPES,
        check_permissions=lambda user: user.has_perm('moonstuff.add_trackingcharacter'),
        is_character_added=_is_character_added,
        is_character_added_annotation=Exists(
            TrackingCharacter.objects
            .filter(character_id=OuterRef('pk'))
        ),
        get_users_with_perms=_users_with_perms,
    ),
])



# Generated by Django 4.2.18 on 2025-02-26 14:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('charlink', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='AppSettings',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('app_name', models.CharField(max_length=255, unique=True)),
                ('ignored', models.BooleanField(default=False)),
                ('default_selection', models.BooleanField(default=True)),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.AlterModelOptions(
            name='general',
            options={'default_permissions': (), 'managed': False, 'permissions': (('view_corp', 'Can view linked character of members of their corporation.'), ('view_alliance', 'Can view linked character of members of their alliance.'), ('view_state', 'Can view linked character of members of their auth state.'), ('view_admin', 'Can view CharLink Admin page.'))},
        ),
    ]



from django.core.management.base import BaseCommand

from charlink.models import AppSettings
from charlink.app_imports import import_apps


class Command(BaseCommand):
    help = 'Reset login default selections'

    def handle(self, *args, **options):
        app_imports = import_apps()

        self.stdout.write('Resetting login options default selections...')

        for app_import in app_imports.values():
            for login_import in app_import.imports:
                AppSettings.objects.update_or_create(
                    app_name=login_import.get_query_id(),
                    defaults={
                        'default_selection': login_import.default_initial_selection,
                    }
                )

        self.stdout.write(self.style.SUCCESS('Reset done!'))



from unittest.mock import patch, Mock

from django.test import TestCase, RequestFactory
from django.urls import reverse
from django.contrib.messages import get_messages, DEFAULT_LEVELS
from django.db.models import OuterRef, Exists

from allianceauth.authentication.models import CharacterOwnership

from app_utils.testdata_factories import UserMainFactory, EveCorporationInfoFactory, EveCharacterFactory

from charlink.views import get_navbar_elements, dashboard_login
from charlink.imports.memberaudit import app_import as memberaudit_import
from charlink.imports.miningtaxes import app_import as miningtaxes_import
from charlink.imports.corptools import _corp_perms
from charlink.app_imports.utils import AppImport, LoginImport
from charlink.app_imports import import_apps
from charlink.models import AppSettings


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestGetNavbarElements(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.permuser = UserMainFactory(permissions=['charlink.view_corp'])
        cls.nopermuser = UserMainFactory()

    def test_with_perm(self):
        res = get_navbar_elements(self.permuser)

        self.assertIn('available', res)
        self.assertTrue(res['is_auditor'])

        self.assertIn('available', res)
        self.assertGreater(len(res['available']), 0)

        self.assertIn('available_apps', res)
        self.assertGreater(len(res['available_apps']), 0)

    def test_without_perm(self):
        res = get_navbar_elements(self.nopermuser)

        self.assertIn('available', res)
        self.assertFalse(res['is_auditor'])

        self.assertIn('available', res)
        self.assertEqual(len(res['available']), 0)

        self.assertIn('available_apps', res)
        self.assertEqual(len(res['available_apps']), 0)


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestDashboardLogin(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(
            permissions=[
                'memberaudit.basic_access',
                'miningtaxes.basic_access',
            ]
        )

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        request_factory = RequestFactory()
        cls.request = request_factory.get('/fake')
        cls.request.user = cls.user

        cls.form_contents = [
            '''<div class="mb-3">
                <div class="form-check">
                    <input type="checkbox" name="charlink-allianceauth.authentication_default" class="form-check-input" disabled id="id_charlink-allianceauth.authentication_default" checked>
                    <label class="form-check-label" for="id_charlink-allianceauth.authentication_default">Add Character (default)</label>
                </div>
            </div>''',
            '''<div class="mb-3">
                <div class="form-check">
                    <input type="checkbox" name="charlink-memberaudit_default" class="form-check-input" id="id_charlink-memberaudit_default" checked>
                    <label class="form-check-label" for="id_charlink-memberaudit_default">Member Audit</label>
                </div>
            </div>''',
            '''<div class="mb-3">
                <div class="form-check">
                    <input type="checkbox" name="charlink-miningtaxes_default" class="form-check-input" id="id_charlink-miningtaxes_default" checked>
                    <label class="form-check-label" for="id_charlink-miningtaxes_default">Mining Taxes</label>
                </div>
            </div>''',
            '''<div class="mb-3">
                <div class="form-check">
                    <input type="checkbox" name="charlink-testauth.testapp_default" class="form-check-input" id="id_charlink-testauth.testapp_default" checked>
                    <label class="form-check-label" for="id_charlink-testauth.testapp_default">TestApp</label>
                </div>
            </div>''',
            '''<div class="mb-3">
                <div class="form-check">
                    <input type="checkbox" name="charlink-testauth.testapp_import2" class="form-check-input" id="id_charlink-testauth.testapp_import2" checked>
                    <label class="form-check-label" for="id_charlink-testauth.testapp_import2">TestApp2</label>
                </div>
            </div>''',
        ]

    def test_ok(self):
        res = dashboard_login(self.request)
        for content in self.form_contents:
            self.assertInHTML(content, res)


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestDashboardPost(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(
            permissions=[
                'corputils.add_corpstats',
                'memberaudit.basic_access',
                'miningtaxes.basic_access',
                'moonmining.add_refinery_owner',
                'moonmining.basic_access',
                'corpstats.add_corpstat',
            ]
        )

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.form_data = {
            'charlink-allianceauth.corputils_default': ['on'],
            'charlink-miningtaxes_default': ['on'],
            'charlink-moonmining_default': ['on'],
            'charlink-corpstats_default': ['on'],
        }

    def test_post_ok(self):
        self.client.force_login(self.user)

        res = self.client.post(reverse('charlink:dashboard_post'), self.form_data)

        self.assertRedirects(
            res,
            reverse('charlink:login'),
            fetch_redirect_response=False
        )

        session = self.client.session

        self.assertIn('charlink', session)
        self.assertIn('scopes', session['charlink'])
        self.assertIn('imports', session['charlink'])

        converted_imports = [list(x) for x in session['charlink']['imports']]

        self.assertIn(['allianceauth.authentication', 'default'], converted_imports)
        self.assertIn(['allianceauth.corputils', 'default'], converted_imports)
        self.assertNotIn(['memberaudit', 'default'], converted_imports)
        self.assertIn(['miningtaxes', 'default'], converted_imports)
        self.assertIn(['moonmining', 'default'], converted_imports)
        self.assertIn(['corpstats', 'default'], converted_imports)
        self.assertEqual(len(session['charlink']['imports']), 5)

    def test_get(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:dashboard_post'))

        self.assertRedirects(res, reverse('charlink:index'))

        messages = list(get_messages(res.wsgi_request))

        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].level, DEFAULT_LEVELS['ERROR'])
        self.assertEqual(messages[0].message, 'Invalid request')

    # force form invalid
    @patch('charlink.forms.LinkForm.is_valid')
    def test_form_invalid(self, mock_is_valid):
        mock_is_valid.return_value = False

        self.client.force_login(self.user)
        res = self.client.post(reverse('charlink:dashboard_post'), self.form_data)

        self.assertRedirects(res, reverse('authentication:dashboard'))

        messages = list(get_messages(res.wsgi_request))

        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].level, DEFAULT_LEVELS['ERROR'])
        self.assertEqual(messages[0].message, 'Invalid form data')


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestIndex(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(
            permissions=[
                'corputils.add_corpstats',
                'memberaudit.basic_access',
                'miningtaxes.basic_access',
                'moonmining.add_refinery_owner',
                'moonmining.basic_access',
                'corpstats.add_corpstat',
            ]
        )

        cls.form_data = {
            'allianceauth.corputils_default': ['on'],
            'miningtaxes_default': ['on'],
            'moonmining_default': ['on'],
            'corpstats_default': ['on'],
        }

    def test_get(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:index'))

        self.assertEqual(res.status_code, 200)
        self.assertIn('form', res.context)
        self.assertIn('characters_added', res.context)

    def test_post_ok(self):
        self.client.force_login(self.user)

        res = self.client.post(reverse('charlink:index'), self.form_data)

        self.assertRedirects(
            res,
            reverse('charlink:login'),
            fetch_redirect_response=False
        )

        session = self.client.session

        self.assertIn('charlink', session)
        self.assertIn('scopes', session['charlink'])
        self.assertIn('imports', session['charlink'])

        converted_imports = [list(x) for x in session['charlink']['imports']]

        self.assertIn(['allianceauth.authentication', 'default'], converted_imports)
        self.assertIn(['allianceauth.corputils', 'default'], converted_imports)
        self.assertNotIn(['memberaudit', 'default'], converted_imports)
        self.assertIn(['miningtaxes', 'default'], converted_imports)
        self.assertIn(['moonmining', 'default'], converted_imports)
        self.assertIn(['corpstats', 'default'], converted_imports)
        self.assertEqual(len(session['charlink']['imports']), 5)

    # form always valid
    # def test_post_wrong_data(self):
    #     self.client.force_login(self.user)

    #     res = self.client.post(reverse('charlink:index'), {'allianceauth.authentication:': '5'})

    #     self.assertEqual(res.status_code, 200)
    #     self.assertIn('form', res.context)
    #     self.assertIn('characters_added', res.context)

    # force form invalid
    @patch('charlink.forms.LinkForm.is_valid')
    def test_form_invalid(self, mock_is_valid):
        mock_is_valid.return_value = False

        self.client.force_login(self.user)
        res = self.client.post(reverse('charlink:index'), self.form_data)

        self.assertEqual(res.status_code, 200)
        self.assertIn('form', res.context)
        self.assertIn('characters_added', res.context)


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestLoginView(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.scopes = list(set(memberaudit_import.imports[0].scopes + miningtaxes_import.imports[0].scopes))
        cls.user = UserMainFactory(
            permissions=[
                'memberaudit.basic_access',
                'miningtaxes.basic_access',
            ],
            main_character__scopes=cls.scopes
        )
        cls.token = cls.user.token_set.first()

    @patch('charlink.imports.miningtaxes._add_character_basic')
    @patch('charlink.imports.memberaudit._add_character')
    @patch('charlink.views.import_apps')
    @patch('charlink.decorators.token_required')
    def test_ok(self, mock_token_required, mock_import_apps, mock_memberaudit_add_character, mock_miningtaxes_add_character):
        import_apps()
        session = self.client.session
        session['charlink'] = {
            'scopes': self.scopes,
            'imports': [
                ('memberaudit', 'default'),
                ('miningtaxes', 'default'),
                ('allianceauth.authentication', 'default'),
            ],
        }
        session.save()

        def fake_decorator(f):
            def fake_wrapper(request, *args, **kwargs):
                return f(request, self.token, *args, **kwargs)
            return fake_wrapper

        mock_token_required.return_value = fake_decorator

        mock_import_apps.return_value = {
            'memberaudit': AppImport('memberaudit', [
                LoginImport(
                    app_label='memberaudit',
                    unique_id='default',
                    field_label='Member Audit',
                    add_character=lambda requets, token: None,
                    scopes=memberaudit_import.imports[0].scopes,
                    check_permissions=memberaudit_import.imports[0].check_permissions,
                    is_character_added=memberaudit_import.imports[0].is_character_added,
                    is_character_added_annotation=memberaudit_import.imports[0].is_character_added_annotation,
                    get_users_with_perms=memberaudit_import.imports[0].get_users_with_perms,
                )
            ]),
            'miningtaxes': AppImport('miningtaxes', [
                LoginImport(
                    app_label='miningtaxes',
                    unique_id='default',
                    field_label='Mining Taxes',
                    add_character=Mock(side_effect=Exception('test')),
                    scopes=miningtaxes_import.imports[0].scopes,
                    check_permissions=miningtaxes_import.imports[0].check_permissions,
                    is_character_added=miningtaxes_import.imports[0].is_character_added,
                    is_character_added_annotation=miningtaxes_import.imports[0].is_character_added_annotation,
                    get_users_with_perms=miningtaxes_import.imports[0].get_users_with_perms,
                )
            ]),
            'allianceauth.authentication': AppImport('allianceauth.authentication', [
                LoginImport(
                    app_label='allianceauth.authentication',
                    unique_id='default',
                    field_label='Add Character (default)',
                    add_character=lambda request, token: None,
                    scopes=['publicData'],
                    check_permissions=lambda user: True,
                    is_character_added=lambda character: CharacterOwnership.objects.filter(character=character).exists(),
                    is_character_added_annotation=Exists(CharacterOwnership.objects.filter(character_id=OuterRef('pk'))),
                    get_users_with_perms=lambda: None,
                )
            ]),
        }

        mock_memberaudit_add_character.return_value = None
        mock_miningtaxes_add_character.side_effect = Exception('test')

        self.client.force_login(self.user)
        res = self.client.get(reverse('charlink:login'))

        messages = list(get_messages(res.wsgi_request))

        self.assertEqual(len(messages), 2)

        sorted_messages = sorted(messages, key=lambda x: x.level)
        self.assertEqual(sorted_messages[0].level, DEFAULT_LEVELS['SUCCESS'])
        self.assertEqual(sorted_messages[1].level, DEFAULT_LEVELS['ERROR'])

    @patch('charlink.imports.miningtaxes._add_character_basic')
    @patch('charlink.imports.memberaudit._add_character')
    @patch('charlink.views.import_apps')
    @patch('charlink.decorators.token_required')
    def test_ignore(self, mock_token_required, mock_import_apps, mock_memberaudit_add_character, mock_miningtaxes_add_character):
        import_apps()
        AppSettings.objects.filter(app_name__startswith='miningtaxes').update(ignored=True)

        session = self.client.session
        session['charlink'] = {
            'scopes': self.scopes,
            'imports': [
                ('memberaudit', 'default'),
                ('miningtaxes', 'default'),
                ('allianceauth.authentication', 'default'),
            ],
        }
        session.save()

        def fake_decorator(f):
            def fake_wrapper(request, *args, **kwargs):
                return f(request, self.token, *args, **kwargs)
            return fake_wrapper

        mock_token_required.return_value = fake_decorator

        mock_import_apps.return_value = {
            'memberaudit': AppImport('memberaudit', [
                LoginImport(
                    app_label='memberaudit',
                    unique_id='default',
                    field_label='Member Audit',
                    add_character=lambda requets, token: None,
                    scopes=memberaudit_import.imports[0].scopes,
                    check_permissions=memberaudit_import.imports[0].check_permissions,
                    is_character_added=memberaudit_import.imports[0].is_character_added,
                    is_character_added_annotation=memberaudit_import.imports[0].is_character_added_annotation,
                    get_users_with_perms=memberaudit_import.imports[0].get_users_with_perms,
                )
            ]),
            'miningtaxes': AppImport('miningtaxes', [
                LoginImport(
                    app_label='miningtaxes',
                    unique_id='default',
                    field_label='Mining Taxes',
                    add_character=Mock(side_effect=Exception('test')),
                    scopes=miningtaxes_import.imports[0].scopes,
                    check_permissions=miningtaxes_import.imports[0].check_permissions,
                    is_character_added=miningtaxes_import.imports[0].is_character_added,
                    is_character_added_annotation=miningtaxes_import.imports[0].is_character_added_annotation,
                    get_users_with_perms=miningtaxes_import.imports[0].get_users_with_perms,
                )
            ]),
            'allianceauth.authentication': AppImport('allianceauth.authentication', [
                LoginImport(
                    app_label='allianceauth.authentication',
                    unique_id='default',
                    field_label='Add Character (default)',
                    add_character=lambda request, token: None,
                    scopes=['publicData'],
                    check_permissions=lambda user: True,
                    is_character_added=lambda character: CharacterOwnership.objects.filter(character=character).exists(),
                    is_character_added_annotation=Exists(CharacterOwnership.objects.filter(character_id=OuterRef('pk'))),
                    get_users_with_perms=lambda: None,
                )
            ]),
        }

        mock_memberaudit_add_character.return_value = None
        mock_miningtaxes_add_character.return_value = None

        self.client.force_login(self.user)
        res = self.client.get(reverse('charlink:login'))

        messages = list(get_messages(res.wsgi_request))

        self.assertEqual(len(messages), 1)

        sorted_messages = sorted(messages, key=lambda x: x.level)
        self.assertEqual(sorted_messages[0].level, DEFAULT_LEVELS['SUCCESS'])


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestAudit(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=['charlink.view_corp'])
        cls.corp = cls.user.profile.main_character.corporation
        cls.corp2 = EveCorporationInfoFactory()
        cls.char2 = EveCharacterFactory(corporation=cls.corp2)
        cls.user2 = UserMainFactory(main_character__character=cls.char2)

    def test_ok(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_corp', args=[self.corp.corporation_id]))

        self.assertEqual(res.status_code, 200)
        self.assertIn('selected', res.context)

    def test_no_perm(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_corp', args=[self.corp2.corporation_id]))

        self.assertNotEqual(res.status_code, 200)


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestSearch(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=['charlink.view_corp'])
        cls.main_char = cls.user.profile.main_character

        cls.user2 = UserMainFactory()
        cls.main_char2 = cls.user2.profile.main_character

    def test_ok(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:search'), {'search_string': self.main_char.character_name})

        self.assertEqual(res.status_code, 200)
        self.assertIn('search_string', res.context)
        self.assertIn('characters', res.context)
        self.assertEqual(len(res.context['characters']), 1)

    def test_not_found(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:search'), {'search_string': self.main_char2.character_name})

        self.assertEqual(res.status_code, 200)
        self.assertIn('search_string', res.context)
        self.assertIn('characters', res.context)
        self.assertEqual(len(res.context['characters']), 0)

    def test_missing_string(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:search'))

        self.assertRedirects(res, reverse('charlink:index'))


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestAuditUser(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=['charlink.view_corp'])

        char2 = EveCharacterFactory(corporation=cls.user.profile.main_character.corporation)
        cls.user2 = UserMainFactory(main_character__character=char2)

        cls.user_ext = UserMainFactory()

    def test_ok(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_user', args=[self.user2.pk]))

        self.assertEqual(res.status_code, 200)
        self.assertIn('characters_added', res.context)

    def test_no_perm(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_user', args=[self.user_ext.pk]))

        self.assertNotEqual(res.status_code, 200)


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestAuditApp(TestCase):

    @classmethod
    def setUpTestData(cls):
        permissions = ["memberaudit.basic_access", "moonmining.add_refinery_owner", "moonmining.basic_access"]
        cls.user = UserMainFactory(
            permissions=[
                'charlink.view_corp',
                'corptools.view_characteraudit',
                *permissions,
                *_corp_perms,
            ]
        )
        char2, char3 = EveCharacterFactory.create_batch(
            2,
            corporation=cls.user.profile.main_character.corporation
        )
        cls.user2 = UserMainFactory(
            permissions=permissions,
            main_character__character=char2
        )
        cls.random_char = EveCharacterFactory(corporation=cls.user.profile.main_character.corporation)
        cls.no_perm_user = UserMainFactory(
            permissions=['charlink.view_corp'],
            main_character__character=char3
        )

    def test_ok(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_app', args=['memberaudit']))

        self.assertEqual(res.status_code, 200)
        self.assertIn('app', res.context)
        self.assertIn('logins', res.context)
        self.assertEqual(len(res.context['logins']), 1)
        self.assertEqual(len(list(res.context['logins'].values())[0]), 2)

    def test_app_empty_perms(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_app', args=['allianceauth.authentication']))

        self.assertEqual(res.status_code, 200)
        self.assertIn('logins', res.context)
        self.assertEqual(len(res.context['logins']), 1)
        self.assertEqual(len(list(res.context['logins'].values())[0]), 3)
        self.assertIn('app', res.context)

    def test_missing_app(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_app', args=['invalid_app']))

        self.assertEqual(res.status_code, 404)

    def test_no_app_perm(self):
        self.client.force_login(self.no_perm_user)

        res = self.client.get(reverse('charlink:audit_app', args=['memberaudit']))

        self.assertNotEqual(res.status_code, 200)

    def test_multiple_app_perms(self):
        self.client.force_login(self.user)

        extra_char = EveCharacterFactory(corporation=self.user.profile.main_character.corporation)
        UserMainFactory(
            permissions=["moonmining.add_refinery_owner", "moonmining.basic_access"],
            main_character__character=extra_char
        )

        res = self.client.get(reverse('charlink:audit_app', args=['moonmining']))

        self.assertEqual(res.status_code, 200)
        self.assertIn('logins', res.context)
        self.assertEqual(len(res.context['logins']), 1)
        self.assertEqual(len(list(res.context['logins'].values())[0]), 3)
        self.assertIn('app', res.context)

    def test_app_with_multiple_logins(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:audit_app', args=['corptools']))

        self.assertEqual(res.status_code, 200)
        self.assertIn('logins', res.context)
        self.assertEqual(len(res.context['logins']), 2)


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestAdminImportedApps(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.admin = UserMainFactory(is_superuser=True)
        cls.user = UserMainFactory()

    @patch('charlink.views.import_apps')
    @patch('charlink.views.get_duplicated_apps')
    @patch('charlink.views.get_failed_to_import')
    @patch('charlink.views.get_no_import')
    def test_admin(self, mock_no_import, mock_failed_to_import, mock_duplicated_apps, mock_import_apps):
        mock_no_import.return_value = []
        mock_failed_to_import.return_value = {}
        mock_duplicated_apps.return_value = set()
        mock_import_apps.return_value = {}

        self.client.force_login(self.admin)

        res = self.client.get(reverse('charlink:admin_imported_apps'))

        self.assertEqual(res.status_code, 200)
        self.assertIn('imported_apps', res.context)
        self.assertIn('duplicated_apps', res.context)
        self.assertIn('failed_to_import', res.context)
        self.assertIn('no_import', res.context)

    def test_no_admin(self):
        self.client.force_login(self.user)

        res = self.client.get(reverse('charlink:admin_imported_apps'))

        self.assertNotEqual(res.status_code, 200)


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestToggleAppVisible(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=['charlink.view_admin'])

    def test_toggle_authentication_login(self):
        self.client.force_login(self.user)

        app_name = 'allianceauth.authentication_default'

        app_settings = AppSettings.objects.create(
            app_name=app_name,
            default_selection=True
        )

        self.assertFalse(app_settings.ignored)

        res = self.client.get(reverse('charlink:toggle_app_visible', args=[app_name]))

        self.assertRedirects(res, reverse('charlink:admin_imported_apps'))

        app_settings.refresh_from_db()
        self.assertFalse(app_settings.ignored)

        messages = list(get_messages(res.wsgi_request))

        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].level, DEFAULT_LEVELS['ERROR'])

    @patch('charlink.app_imports.import_apps')
    def test_toggle_other_login(self, mock_import_apps):
        mock_import_apps.return_value = {
            'testapp': AppImport('testapp', [
                LoginImport(
                    app_label='testapp',
                    unique_id='default',
                    field_label='TestApp',
                    add_character=lambda request, token: None,
                    scopes=['scope1'],
                    check_permissions=lambda user: True,
                    is_character_added=lambda character: False,
                    is_character_added_annotation=Mock(),
                    get_users_with_perms=lambda: None,
                )
            ]),
        }

        self.client.force_login(self.user)

        app_name = 'testapp_default'

        app_settings = AppSettings.objects.create(
            app_name=app_name,
            default_selection=True
        )

        self.assertFalse(app_settings.ignored)

        res = self.client.get(reverse('charlink:toggle_app_visible', args=[app_name]))

        self.assertRedirects(res, reverse('charlink:admin_imported_apps'))

        app_settings.refresh_from_db()
        self.assertTrue(app_settings.ignored)

        messages = list(get_messages(res.wsgi_request))

        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].level, DEFAULT_LEVELS['SUCCESS'])


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestToggleAppDefaultSelection(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=['charlink.view_admin'])

    def test_toggle_authentication_login(self):
        self.client.force_login(self.user)

        app_name = 'allianceauth.authentication_default'

        app_settings = AppSettings.objects.create(
            app_name=app_name,
            default_selection=True
        )

        self.assertTrue(app_settings.default_selection)

        res = self.client.get(reverse('charlink:toggle_app_default_selection', args=[app_name]))

        self.assertRedirects(res, reverse('charlink:admin_imported_apps'))

        app_settings.refresh_from_db()
        self.assertTrue(app_settings.default_selection)

        messages = list(get_messages(res.wsgi_request))

        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].level, DEFAULT_LEVELS['ERROR'])

    @patch('charlink.views.import_apps')
    @patch('charlink.app_imports.import_apps')
    def test_toggle_other_login_default_true(self, mock_views_import_apps, mock_app_imports_import_apps):
        mock_import_apps = {
            'testapp': AppImport('testapp', [
                LoginImport(
                    app_label='testapp',
                    unique_id='default',
                    field_label='TestApp',
                    add_character=lambda request, token: None,
                    scopes=['scope1'],
                    check_permissions=lambda user: True,
                    is_character_added=lambda character: False,
                    is_character_added_annotation=Mock(),
                    get_users_with_perms=lambda: None,
                )
            ]),
        }
        mock_views_import_apps.return_value = mock_import_apps
        mock_app_imports_import_apps.return_value = mock_import_apps

        self.client.force_login(self.user)

        app_name = 'testapp_default'

        app_settings = AppSettings.objects.create(
            app_name=app_name,
            default_selection=True
        )

        res = self.client.get(reverse('charlink:toggle_app_default_selection', args=[app_name]))

        self.assertRedirects(res, reverse('charlink:admin_imported_apps'))

        app_settings.refresh_from_db()
        self.assertFalse(app_settings.default_selection)

        messages = list(get_messages(res.wsgi_request))

        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].level, DEFAULT_LEVELS['SUCCESS'])

    @patch('charlink.views.import_apps')
    @patch('charlink.app_imports.import_apps')
    def test_toggle_other_login_default_false(self, mock_views_import_apps, mock_app_imports_import_apps):
        mock_import_apps = {
            'testapp': AppImport('testapp', [
                LoginImport(
                    app_label='testapp',
                    unique_id='default',
                    field_label='TestApp',
                    add_character=lambda request, token: None,
                    scopes=['scope1'],
                    check_permissions=lambda user: True,
                    is_character_added=lambda character: False,
                    is_character_added_annotation=Mock(),
                    get_users_with_perms=lambda: None,
                    default_initial_selection=False,
                )
            ]),
        }
        mock_views_import_apps.return_value = mock_import_apps
        mock_app_imports_import_apps.return_value = mock_import_apps

        self.client.force_login(self.user)

        app_name = 'testapp_default'

        app_settings = AppSettings.objects.create(
            app_name=app_name,
            default_selection=False
        )

        res = self.client.get(reverse('charlink:toggle_app_default_selection', args=[app_name]))

        self.assertRedirects(res, reverse('charlink:admin_imported_apps'))

        app_settings.refresh_from_db()
        self.assertTrue(app_settings.default_selection)

        messages = list(get_messages(res.wsgi_request))

        self.assertEqual(len(messages), 2)
        self.assertEqual(messages[0].level, DEFAULT_LEVELS['SUCCESS'])
        self.assertEqual(messages[1].level, DEFAULT_LEVELS['WARNING'])



from unittest.mock import patch

from django.test import TestCase

from app_utils.testdata_factories import UserMainFactory

from charlink.imports.moonstuff import _add_character, _is_character_added
from charlink.app_imports import import_apps


class TestAddCharacter(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=['moonstuff.add_trackingcharacter'])

    @patch('charlink.imports.moonstuff.import_extraction_data.delay')
    def test_ok(self, mock_import_extraction_data):
        mock_import_extraction_data.return_value = None

        token = self.user.token_set.first()

        _add_character(None, token)

        self.assertTrue(_is_character_added(self.user.profile.main_character))

        with self.assertRaises(AssertionError):
            _add_character(None, token)

        mock_import_extraction_data.assert_called_once()
        self.assertTrue(_is_character_added(self.user.profile.main_character))


class TestIsCharacterAdded(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=['moonstuff.add_trackingcharacter'])
        cls.character = cls.user.profile.main_character

    @patch('charlink.imports.moonstuff.import_extraction_data.delay')
    def test_ok(self, mock_import_extraction_data):
        mock_import_extraction_data.return_value = None

        self.assertFalse(_is_character_added(self.character))

        token = self.user.token_set.first()

        _add_character(None, token)

        self.assertTrue(_is_character_added(self.character))


class TestCheckPermissions(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.no_perm_user = UserMainFactory()
        cls.perm_user = UserMainFactory(permissions=["moonstuff.add_trackingcharacter"])

    def test_ok(self):
        login_import = import_apps()['moonstuff'].get('default')

        self.assertTrue(login_import.check_permissions(self.perm_user))
        self.assertFalse(login_import.check_permissions(self.no_perm_user))


class TestGetUsersWithPerms(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.no_perm_user = UserMainFactory()
        cls.perm_user = UserMainFactory(permissions=["moonstuff.add_trackingcharacter"])

    def test_ok(self):
        login_import = import_apps()['moonstuff'].get('default')

        users = login_import.get_users_with_perms()
        self.assertEqual(users.count(), 1)
        self.assertEqual(users.first(), self.perm_user)



from django.db.models import Exists, OuterRef
from django.contrib.auth.models import Permission
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.corputils.models import CorpStats

from charlink.app_imports.utils import LoginImport, AppImport

from app_utils.allianceauth import users_with_permission


def _add_character(request, token):
    corp_id = EveCharacter.objects.get(character_id=token.character_id).corporation_id
    try:
        corp = EveCorporationInfo.objects.get(corporation_id=corp_id)
    except EveCorporationInfo.DoesNotExist:
        corp = EveCorporationInfo.objects.create_corporation(corp_id)
    cs = CorpStats.objects.create(token=token, corp=corp)
    cs.update()
    assert cs.pk  # ensure update was successful


def _is_character_added(character: EveCharacter):
    return (
        CorpStats.objects
        .filter(token__character_id=character.character_id)
        .exists()
    )


def _users_with_perms():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label='corputils',
            codename='add_corpstats'
        )
    )


app_import = AppImport('allianceauth.corputils', [
    LoginImport(
        app_label='allianceauth.corputils',
        unique_id='default',
        field_label=_('Corporation Stats'),
        add_character=_add_character,
        scopes=['esi-corporations.read_corporation_membership.v1'],
        check_permissions=lambda user: user.has_perm('corputils.add_corpstats'),
        is_character_added=_is_character_added,
        is_character_added_annotation=Exists(
            CorpStats.objects
            .filter(token__character_id=OuterRef('character_id'))
        ),
        get_users_with_perms=_users_with_perms,
        default_initial_selection=False,
    ),
])



from django.test import TestCase

from charlink.app_imports import import_apps

from marketmanager.views import CHARACTER_SCOPES, CORPORATION_SCOPES

from app_utils.testdata_factories import UserMainFactory, EveCharacterFactory
from app_utils.testing import add_character_to_user


class TestIsCharacterAdded(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()
        cls.main_character = cls.user.profile.main_character

        cls.login_char = EveCharacterFactory()
        add_character_to_user(cls.user, cls.login_char, scopes=CHARACTER_SCOPES)
        cls.login_corp = EveCharacterFactory()
        add_character_to_user(cls.user, cls.login_corp, scopes=CORPORATION_SCOPES)

    def test_ok_character_login(self):
        app_import = import_apps()['marketmanager']

        self.assertFalse(app_import.get('character').is_character_added(self.main_character))
        self.assertTrue(app_import.get('character').is_character_added(self.login_char))
        self.assertFalse(app_import.get('character').is_character_added(self.login_corp))

    def test_ok_corporation_login(self):
        app_import = import_apps()['marketmanager']

        self.assertFalse(app_import.get('corporation').is_character_added(self.main_character))
        self.assertFalse(app_import.get('corporation').is_character_added(self.login_char))
        self.assertTrue(app_import.get('corporation').is_character_added(self.login_corp))


class TestAddCharacter(TestCase):

    def test_ok(self):
        app_import = import_apps()['marketmanager']

        self.assertIsNone(app_import.get('character').add_character(None, None))
        self.assertIsNone(app_import.get('corporation').add_character(None, None))


class TestCheckPermissions(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.basic_market_browser_user = UserMainFactory(permissions=['marketmanager.basic_market_browser'])
        cls.no_perm_user = UserMainFactory()

    def test_ok(self):
        app_import = import_apps()['marketmanager']

        self.assertTrue(app_import.get('character').check_permissions(self.basic_market_browser_user))
        self.assertTrue(app_import.get('corporation').check_permissions(self.basic_market_browser_user))
        self.assertFalse(app_import.get('character').check_permissions(self.no_perm_user))
        self.assertFalse(app_import.get('corporation').check_permissions(self.no_perm_user))


class TestGetUsersWithPerms(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.basic_market_browser_user = UserMainFactory(permissions=['marketmanager.basic_market_browser'])
        cls.no_perm_user = UserMainFactory()

    def test_ok(self):
        app_import = import_apps()['marketmanager']

        self.assertQuerysetEqual(
            app_import.get('character').get_users_with_perms(),
            [self.basic_market_browser_user.pk],
            ordered=False,
            transform=lambda x: x.pk
        )
        self.assertQuerysetEqual(
            app_import.get('corporation').get_users_with_perms(),
            [self.basic_market_browser_user.pk],
            ordered=False,
            transform=lambda x: x.pk
        )



from unittest.mock import patch

from django.test import TestCase, RequestFactory
from django.contrib.messages.storage.fallback import FallbackStorage

from app_utils.testdata_factories import UserMainFactory

from charlink.imports.miningtaxes import _add_character_basic, _is_character_added_basic, _add_character_admin, _is_character_added_admin
from charlink.app_imports import import_apps

from miningtaxes.models import Character, AdminCharacter


class TestAddCharacter(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=["miningtaxes.basic_access"])
        cls.character = cls.user.profile.main_character

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()

    @patch('miningtaxes.tasks.update_character.delay')
    def test_ok_basic(self, mock_update_character):
        mock_update_character.return_value = None

        token = self.user.token_set.first()

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages

        _add_character_basic(request, token)

        mock_update_character.assert_called_once()
        self.assertTrue(_is_character_added_basic(self.character))

    @patch('miningtaxes.tasks.update_admin_character.delay')
    def test_ok_admin(self, mock_update_character):
        mock_update_character.return_value = None

        token = self.user.token_set.first()

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages

        _add_character_admin(request, token)

        mock_update_character.assert_called_once()
        self.assertTrue(_is_character_added_admin(self.character))


class TestIsCharacterAdded(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=["miningtaxes.basic_access"])
        cls.character = cls.user.profile.main_character

    def test_ok_basic(self):
        self.assertFalse(_is_character_added_basic(self.character))
        Character.objects.create(eve_character=self.character)
        self.assertTrue(_is_character_added_basic(self.character))

    def test_ok_admin(self):
        self.assertFalse(_is_character_added_admin(self.character))
        AdminCharacter.objects.create(eve_character=self.character)
        self.assertTrue(_is_character_added_admin(self.character))


class TestCheckPermissions(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.no_perm_user = UserMainFactory()
        cls.perm_user_basic = UserMainFactory(permissions=["miningtaxes.basic_access"])
        cls.perm_user_admin = UserMainFactory(permissions=["miningtaxes.admin_access"])

    def test_ok_basic(self):
        login_import = import_apps()['miningtaxes'].get('default')

        self.assertTrue(login_import.check_permissions(self.perm_user_basic))
        self.assertFalse(login_import.check_permissions(self.no_perm_user))

    def test_ok_admin(self):
        login_import = import_apps()['miningtaxes'].get('admin')

        self.assertTrue(login_import.check_permissions(self.perm_user_admin))
        self.assertFalse(login_import.check_permissions(self.no_perm_user))


class TestGetUsersWithPerms(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.no_perm_user = UserMainFactory()
        cls.perm_user_basic = UserMainFactory(permissions=["miningtaxes.basic_access"])
        cls.perm_user_admin = UserMainFactory(permissions=["miningtaxes.admin_access"])

    def test_ok_basic(self):
        login_import = import_apps()['miningtaxes'].get('default')

        users = login_import.get_users_with_perms()
        self.assertEqual(users.count(), 1)
        self.assertEqual(users.first(), self.perm_user_basic)

    def test_ok_admin(self):
        login_import = import_apps()['miningtaxes'].get('admin')

        users = login_import.get_users_with_perms()
        self.assertEqual(users.count(), 1)
        self.assertEqual(users.first(), self.perm_user_admin)



from io import StringIO
from unittest.mock import Mock, patch

from django.core.management import call_command
from django.test import TestCase

from charlink.app_imports.utils import AppImport, LoginImport
from charlink.models import AppSettings


class CharlinkResetSelectionsCommandTest(TestCase):

    @patch('charlink.management.commands.charlink_reset_selections.import_apps')
    def test_handle_resets_default_selections(self, mock_import_apps):
        mock_import_apps.return_value = {
            'testapp': AppImport('testapp', [
                LoginImport(
                    app_label='testapp',
                    unique_id='default',
                    field_label='TestApp',
                    add_character=lambda request, token: None,
                    scopes=['scope1'],
                    check_permissions=lambda user: True,
                    is_character_added=lambda character: False,
                    is_character_added_annotation=Mock(),
                    get_users_with_perms=lambda: None,
                    default_initial_selection=False,
                ),
                LoginImport(
                    app_label='testapp',
                    unique_id='option2',
                    field_label='TestApp 2',
                    add_character=lambda request, token: None,
                    scopes=['scope1'],
                    check_permissions=lambda user: True,
                    is_character_added=lambda character: False,
                    is_character_added_annotation=Mock(),
                    get_users_with_perms=lambda: None,
                    default_initial_selection=True,
                ),
            ])
        }

        as1 = AppSettings.objects.create(
            app_name='testapp_default',
            default_selection=True
        )
        as2 = AppSettings.objects.create(
            app_name='testapp_option2',
            default_selection=False
        )

        out = StringIO()
        call_command('charlink_reset_selections', stdout=out)

        as1.refresh_from_db()
        as2.refresh_from_db()

        self.assertFalse(as1.default_selection)
        self.assertTrue(as2.default_selection)

        self.assertIn('Resetting login options default selections...', out.getvalue())
        self.assertIn('Reset done!', out.getvalue())



from django.db.models import Exists, OuterRef
from django.contrib.auth.models import Permission, User
from django.utils.translation import gettext_lazy as _

from charlink.app_imports.utils import LoginImport, AppImport

from allianceauth.eveonline.models import EveCharacter

from app_utils.allianceauth import users_with_permission

from esi.models import Token

_scopes_readfleet = ["esi-fleets.read_fleet.v1"]
_scopes_clickfleet = ["esi-location.read_location.v1", "esi-location.read_ship_type.v1", "esi-location.read_online.v1"]


def _is_character_added_readfleet(character: EveCharacter):
    return (
        Token.objects
        .filter(character_id=character.character_id)
        .require_valid()
        .require_scopes(_scopes_readfleet)
        .exists()
    )


def _is_character_added_clickfleet(character: EveCharacter):
    return (
        Token.objects
        .filter(character_id=character.character_id)
        .require_valid()
        .require_scopes(_scopes_clickfleet)
        .exists()
    )


def _check_perms_readfleet(user: User):
    return user.has_perm('afat.manage_afat') or user.has_perm('afat.add_fatlink')


def _users_with_perms_readfleet():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label='afat',
            codename='manage_afat'
        )
    ) | users_with_permission(
        Permission.objects.get(
            content_type__app_label='afat',
            codename='add_fatlink'
        )
    )


def _users_with_perms_clickfleet():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label='afat',
            codename='basic_access'
        )
    )


app_import = AppImport('afat', [
    LoginImport(
        app_label='afat',
        unique_id='readfleet',
        field_label=_('AFAT Read Fleet'),
        add_character=lambda requets, token: None,
        scopes=_scopes_readfleet,
        check_permissions=_check_perms_readfleet,
        is_character_added=_is_character_added_readfleet,
        is_character_added_annotation=Exists(
            Token.objects.all()

            .filter(character_id=OuterRef('character_id'))
            .require_scopes(_scopes_readfleet)
            # .require_valid()
        ),
        get_users_with_perms=_users_with_perms_readfleet,
    ),
    LoginImport(
        app_label='afat',
        unique_id='clickfat',
        field_label=_('AFAT Click Tracking'),
        add_character=lambda request, token: None,
        scopes=_scopes_clickfleet,
        check_permissions=lambda user: user.has_perm('afat.basic_access'),
        is_character_added=_is_character_added_clickfleet,
        is_character_added_annotation=Exists(
            Token.objects.all()
            .filter(character_id=OuterRef('character_id'))
            .require_scopes(_scopes_clickfleet)
            # .require_valid()
        ),
        get_users_with_perms=_users_with_perms_clickfleet,
    )
])



from unittest.mock import patch

from django.test import TestCase, RequestFactory
from django.contrib.messages.storage.fallback import FallbackStorage

from app_utils.testdata_factories import UserMainFactory

from charlink.imports.aa_contacts import (
    _alliance_login,
    _corporation_login,
    _alliance_users_with_perms,
    _corporation_users_with_perms,
    _alliance_check_perms,
    _corporation_check_perms,
    _alliance_is_character_added,
    _corporation_is_character_added
)

from aa_contacts.models import AllianceToken, CorporationToken


class TestAddCharacter(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()
        cls.character = cls.user.profile.main_character
        cls.token = cls.user.token_set.first()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()

    @patch("charlink.imports.aa_contacts.update_alliance_contacts.delay")
    def test_alliance_login_ok(self, mock_update_alliance_contacts):
        mock_update_alliance_contacts.return_value = None

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages
        request.user = self.user

        _alliance_login(request, self.token)

        self.assertTrue(AllianceToken.objects.filter(alliance=self.character.alliance).exists())
        self.assertTrue(mock_update_alliance_contacts.called)

    @patch("charlink.imports.aa_contacts.update_alliance_contacts.delay")
    def test_alliance_login_missing_alliance(self, mock_update_alliance_contacts):
        mock_update_alliance_contacts.return_value = None

        self.character.alliance.delete()

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages
        request.user = self.user

        _alliance_login(request, self.token)

        self.assertTrue(AllianceToken.objects.filter(alliance=self.character.alliance).exists())
        self.assertTrue(mock_update_alliance_contacts.called)

    @patch("charlink.imports.aa_contacts.update_corporation_contacts.delay")
    def test_character_not_in_alliance(self, mock_update_corporation_contacts):
        mock_update_corporation_contacts.return_value = None

        self.character.alliance.delete()
        self.character.alliance_id = None
        self.character.save()

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages
        request.user = self.user

        with self.assertRaises(AssertionError):
            _alliance_login(request, self.token)

        self.assertFalse(AllianceToken.objects.filter(alliance=self.character.alliance).exists())
        self.assertFalse(mock_update_corporation_contacts.called)

    @patch("charlink.imports.aa_contacts.update_corporation_contacts.delay")
    def test_alliance_already_tracked(self, mock_update_corporation_contacts):
        mock_update_corporation_contacts.return_value = None

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages
        request.user = self.user

        AllianceToken.objects.create(alliance=self.character.alliance, token=self.token)
        with self.assertRaises(AssertionError):
            _alliance_login(request, self.token)

        self.assertTrue(AllianceToken.objects.filter(alliance=self.character.alliance).exists())
        self.assertFalse(mock_update_corporation_contacts.called)

    @patch("charlink.imports.aa_contacts.update_corporation_contacts.delay")
    def test_corporation_login_ok(self, mock_update_corporation_contacts):
        mock_update_corporation_contacts.return_value = None

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages
        request.user = self.user

        _corporation_login(request, self.token)

        self.assertTrue(CorporationToken.objects.filter(corporation=self.character.corporation).exists())
        self.assertTrue(mock_update_corporation_contacts.called)

    @patch("charlink.imports.aa_contacts.update_corporation_contacts.delay")
    def test_corporation_login_missing_corporation(self, mock_update_corporation_contacts):
        mock_update_corporation_contacts.return_value = None

        self.character.corporation.delete()

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages
        request.user = self.user

        _corporation_login(request, self.token)

        self.assertTrue(CorporationToken.objects.filter(corporation=self.character.corporation).exists())
        self.assertTrue(mock_update_corporation_contacts.called)

    @patch("charlink.imports.aa_contacts.update_corporation_contacts.delay")
    def test_corporation_already_tracked(self, mock_update_corporation_contacts):
        mock_update_corporation_contacts.return_value = None

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages
        request.user = self.user

        CorporationToken.objects.create(corporation=self.character.corporation, token=self.token)
        with self.assertRaises(AssertionError):
            _corporation_login(request, self.token)

        self.assertTrue(CorporationToken.objects.filter(corporation=self.character.corporation).exists())
        self.assertFalse(mock_update_corporation_contacts.called)


class TestGetUsersWithPerms(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user_no_perm = UserMainFactory()
        cls.user_alliance_perm = UserMainFactory(permissions=["aa_contacts.manage_alliance_contacts"])
        cls.user_corporation_perm = UserMainFactory(permissions=["aa_contacts.manage_corporation_contacts"])

    def test_alliance_users_with_perms(self):
        users = _alliance_users_with_perms()

        self.assertEqual(users.count(), 1)
        self.assertEqual(users.first(), self.user_alliance_perm)

    def test_corporation_users_with_perms(self):
        users = _corporation_users_with_perms()

        self.assertEqual(users.count(), 1)
        self.assertEqual(users.first(), self.user_corporation_perm)


class TestCheckPermissions(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user_no_perm = UserMainFactory()
        cls.user_alliance_perm = UserMainFactory(permissions=["aa_contacts.manage_alliance_contacts"])
        cls.user_corporation_perm = UserMainFactory(permissions=["aa_contacts.manage_corporation_contacts"])

    def test_alliance_check_perms(self):
        self.assertTrue(_alliance_check_perms(self.user_alliance_perm))
        self.assertFalse(_alliance_check_perms(self.user_no_perm))

    def test_corporation_check_perms(self):
        self.assertTrue(_corporation_check_perms(self.user_corporation_perm))
        self.assertFalse(_corporation_check_perms(self.user_no_perm))


class TestIsCharacterAdded(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=["aa_contacts.manage_alliance_contacts"])
        cls.character = cls.user.profile.main_character
        cls.token = cls.user.token_set.first()

    def test_alliance_ok(self):
        self.assertFalse(_alliance_is_character_added(self.character))
        AllianceToken.objects.create(alliance=self.character.alliance, token=self.token)
        self.assertTrue(_alliance_is_character_added(self.character))

    def test_corporation_ok(self):
        self.assertFalse(_corporation_is_character_added(self.character))
        CorporationToken.objects.create(corporation=self.character.corporation, token=self.token)
        self.assertTrue(_corporation_is_character_added(self.character))



from unittest.mock import patch

from django.test import TestCase
from django.contrib.auth.models import User

from app_utils.testdata_factories import UserMainFactory, EveCharacterFactory
from app_utils.testing import add_character_to_user, add_new_token

from charlink.imports.allianceauth.authentication import app_import as auth_import
from charlink.imports.corptools import app_import as corptools_import
from charlink.app_imports import import_apps
from charlink.models import AppSettings, BaseFilter, ComplianceFilter

from corptools.models import CharacterAudit

from esi.models import Token


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestAppSettings(TestCase):

    def test_str_method(self):
        import_apps()

        app_setting = AppSettings.objects.get(app_name='allianceauth.authentication_default')

        self.assertEqual(str(app_setting), str(auth_import.get('default').field_label))


class BaseFilterTestImpl(BaseFilter):
    class Meta:
        app_label = 'charlink'


class TestBaseFilter(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.base_filter = BaseFilterTestImpl(name="test", description="test description")

    def test_str_method(self):
        self.assertEqual(str(self.base_filter), "test: test description")

    def test_process_filter_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.base_filter.process_filter(None)

    def test_audit_filter_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.base_filter.audit_filter(None)


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestComplianceFilter(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(main_character__scopes=corptools_import.get('default').scopes)
        cls.compliance_filter = ComplianceFilter.objects.create(name="compliance test", description="compliance description")

    @patch('charlink.models.ComplianceFilter.audit_filter')
    def test_process_filter(self, mock_audit_filter):
        mock_audit_filter.return_value = {self.user.pk: {'check': True, 'message': 'All good'}}

        res = self.compliance_filter.process_filter(self.user)

        mock_audit_filter.assert_called_once()
        self.assertTrue(res)

    def test_audit_filter_simple(self):
        import_apps()
        self.compliance_filter.selected_apps.add(AppSettings.objects.get(app_name='allianceauth.authentication_default'))
        res = self.compliance_filter.audit_filter(User.objects.filter(pk=self.user.pk))
        self.assertDictEqual(res, {self.user.pk: {'check': True, 'message': 'Meets requirements'}})

    @patch('charlink.imports.corptools.update_character.apply_async')
    def test_audit_filter_with_2_apps(self, mock_update_character):
        mock_update_character.return_value = None

        character = EveCharacterFactory()
        add_character_to_user(self.user, character)

        import_apps()
        self.compliance_filter.selected_apps.add(AppSettings.objects.get(app_name='allianceauth.authentication_default'))
        self.compliance_filter.selected_apps.add(AppSettings.objects.get(app_name='corptools_default'))

        res = self.compliance_filter.audit_filter(User.objects.filter(pk=self.user.pk))
        self.assertDictEqual(res, {self.user.pk: {'check': False, 'message': 'Does not meet requirements'}})

        token2 = add_new_token(self.user, character, scopes=corptools_import.get('default').scopes)
        corptools_import.get('default').add_character(None, token2)

        res = self.compliance_filter.audit_filter(User.objects.filter(pk=self.user.pk))
        self.assertDictEqual(res, {self.user.pk: {'check': False, 'message': 'Does not meet requirements'}})

        main_token = Token.objects.get(user=self.user, character_id=self.user.profile.main_character.character_id)
        corptools_import.get('default').add_character(None, main_token)

        CharacterAudit.objects.update(active=True)

        res = self.compliance_filter.audit_filter(User.objects.filter(pk=self.user.pk))
        self.assertDictEqual(res, {self.user.pk: {'check': True, 'message': 'Meets requirements'}})

    def test_audit_filter_removed_app(self):
        new_as = AppSettings.objects.create(app_name='non_existent_app', default_selection=True)
        self.compliance_filter.selected_apps.add(new_as)

        res = self.compliance_filter.audit_filter(User.objects.filter(pk=self.user.pk))
        self.assertFalse(res[self.user.pk]['check'])
        self.assertEqual(res[self.user.pk]['message'], '')



from importlib import import_module
from unittest.mock import patch

from django.test import TestCase
from django.db.models import Q

from allianceauth.tests.auth_utils import AuthUtils

from app_utils.testdata_factories import UserMainFactory

from charlink.app_imports import import_apps, get_duplicated_apps, get_failed_to_import, get_no_import
from charlink.imports.corptools import _corp_perms
from charlink.models import AppSettings


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestImportApps(TestCase):

    @patch('charlink.app_imports.import_module', wraps=import_module)
    @patch('charlink.app_imports._imported', False)
    @patch('charlink.app_imports._duplicated_apps', set())
    @patch('charlink.app_imports._supported_apps', {})
    def test_not_imported(self, mock_import_module):
        imported_apps = import_apps()
        failed = get_failed_to_import()
        no_import = get_no_import()
        duplicated = get_duplicated_apps()

        self.assertTrue(mock_import_module.called)
        self.assertIn('allianceauth.authentication', imported_apps)
        self.assertIn('allianceauth.corputils', imported_apps)
        self.assertIn('testauth.testapp', imported_apps)
        self.assertNotIn('fakeapp', imported_apps)
        self.assertIn('testauth.testapp.charlink_hook_invalid', failed)
        self.assertNotIn('fakeapp2', imported_apps)
        self.assertIn('testauth.testapp.charlink_hook_no_import', failed)
        self.assertNotIn('allianceauth', imported_apps)
        self.assertNotIn('allianceauth.eveonline', imported_apps)
        self.assertIn('testauth.testapp.charlink_hook_no_import', failed)
        mock_import_module.assert_any_call('charlink.imports.allianceauth.eveonline')

        self.assertSetEqual(
            {
                'allianceauth.corputils_default',
                'structures_default',
                'moonstuff_default',
                'testauth.testapp_default',
                'marketmanager_corporation',
                'testauth.testapp_import2',
                'aa_contacts_corporation',
                'miningtaxes_default',
                'corpstats_default',
                'marketmanager_character',
                'miningtaxes_admin',
                'moonmining_default',
                'afat_clickfat',
                'corptools_default',
                'memberaudit_default',
                'allianceauth.authentication_default',
                'afat_readfleet',
                'corptools_structures',
                'aa_contacts_alliance',
            },
            set(AppSettings.objects.values_list('app_name', flat=True))
        )

        self.assertSetEqual({'testauth.testapp_duplicate'}, duplicated)
        self.assertIn('charlink', no_import)

    @patch('charlink.app_imports.import_module', wraps=import_module)
    def test_imported(self, mock_import_module):
        import_apps()
        mock_import_module.reset_mock()
        imported_apps = import_apps()
        self.assertFalse(mock_import_module.called)
        self.assertIn('allianceauth.authentication', imported_apps)
        self.assertIn('allianceauth.corputils', imported_apps)
        self.assertIn('testauth.testapp', imported_apps)
        self.assertNotIn('allianceauth', imported_apps)
        self.assertNotIn('allianceauth.eveonline', imported_apps)

    def test_supported_apps_default(self):
        user = UserMainFactory()
        main_char = user.profile.main_character

        add_char = import_apps()['allianceauth.authentication']
        self.assertIsNone(add_char.imports[0].add_character(None, None))
        self.assertTrue(add_char.imports[0].is_character_added(main_char))
        self.assertTrue(add_char.imports[0].check_permissions(user))
        self.assertEqual(add_char.imports[0].get_users_with_perms().count(), 1)

    def test_ignore_duplicate_imports(self):
        imported_apps = import_apps()
        self.assertNotIn('testauth.testapp_duplicate', imported_apps)
        self.assertSetEqual({'testauth.testapp_duplicate'}, get_duplicated_apps())

    @patch('charlink.app_imports.import_module', wraps=import_module)
    def test_get_duplicated_apps_imports_apps(self, mock_import_module):
        get_duplicated_apps()
        self.assertTrue(mock_import_module.called)


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestLoginImport(TestCase):

    def test_get_query_id(self):
        login_import = import_apps()['allianceauth.authentication'].get('default')
        self.assertEqual(login_import.get_query_id(), 'allianceauth.authentication_default')

    def test_hash(self):
        login_import = import_apps()['allianceauth.authentication'].get('default')
        self.assertEqual(hash(login_import), hash('allianceauth.authentication_default'))

    def test_is_ignored(self):
        marketmanager = import_apps()['marketmanager']

        AppSettings.objects.filter(app_name='marketmanager_corporation').update(ignored=True)

        corporation_import = marketmanager.get('corporation')
        character_import = marketmanager.get('character')

        self.assertTrue(corporation_import.is_ignored)
        self.assertFalse(character_import.is_ignored)

    def test_default_selection(self):
        marketmanager = import_apps()['marketmanager']

        AppSettings.objects.filter(app_name='marketmanager_corporation').update(default_selection=False)

        corporation_import = marketmanager.get('corporation')
        character_import = marketmanager.get('character')

        self.assertFalse(corporation_import.default_selection)
        self.assertTrue(character_import.default_selection)


@patch('charlink.app_imports._imported', False)
@patch('charlink.app_imports._duplicated_apps', set())
@patch('charlink.app_imports._supported_apps', {})
@patch('charlink.app_imports._failed_to_import', {})
@patch('charlink.app_imports._no_import', [])
class TestAppImport(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()

    def test_get_form_fields(self):
        imported_apps = import_apps()
        user = AuthUtils.add_permissions_to_user_by_name(['corptools.view_characteraudit', "marketmanager.basic_market_browser", *_corp_perms], self.user)
        app_import = imported_apps['allianceauth.authentication']
        form_fields_auth = app_import.get_form_fields(user)
        self.assertEqual(len(form_fields_auth), 1)
        self.assertIn('allianceauth.authentication_default', form_fields_auth)

        marketmanager = imported_apps['marketmanager']
        form_fields_marketmanager = marketmanager.get_form_fields(user)
        self.assertEqual(len(form_fields_marketmanager), 2)
        self.assertIn('marketmanager_corporation', form_fields_marketmanager)
        self.assertIn('marketmanager_character', form_fields_marketmanager)

        corptools = imported_apps['corptools']
        form_fields_corptools = corptools.get_form_fields(user)
        self.assertEqual(len(form_fields_corptools), 2)
        self.assertIn('corptools_default', form_fields_corptools)
        self.assertIn('corptools_structures', form_fields_corptools)

    def test_get_form_fields_ignore(self):
        user = AuthUtils.add_permissions_to_user_by_name(['corptools.view_characteraudit', "marketmanager.basic_market_browser", *_corp_perms], self.user)
        app_import = import_apps()['allianceauth.authentication']

        AppSettings.objects.filter(
            Q(app_name='marketmanager_corporation') | Q(app_name__startswith='corptools')
        ).update(ignored=True)

        form_fields = app_import.get_form_fields(user)
        self.assertEqual(len(form_fields), 1)
        self.assertIn('allianceauth.authentication_default', form_fields)

        marketmanager = import_apps()['marketmanager']
        form_fields = marketmanager.get_form_fields(user)
        self.assertEqual(len(form_fields), 1)
        self.assertIn('marketmanager_character', form_fields)
        self.assertNotIn('marketmanager_corporation', form_fields)

        corptools = import_apps()['corptools']
        form_fields = corptools.get_form_fields(user)
        self.assertEqual(len(form_fields), 0)

    def test_get_imports_with_perms(self):
        app_import = import_apps()['allianceauth.authentication']
        imports = app_import.get_imports_with_perms(self.user)
        self.assertEqual(len(imports.imports), 1)

        user_corp = UserMainFactory(permissions=_corp_perms)
        user_charaudit = UserMainFactory(permissions=['corptools.view_characteraudit'])
        user_both = UserMainFactory(permissions=['corptools.view_characteraudit', *_corp_perms])

        corptools_import = import_apps()['corptools']

        imports_corp = corptools_import.get_imports_with_perms(user_corp)
        self.assertEqual(len(imports_corp.imports), 1)
        self.assertEqual(imports_corp.imports[0].unique_id, 'structures')

        imports_charaudit = corptools_import.get_imports_with_perms(user_charaudit)
        self.assertEqual(len(imports_charaudit.imports), 1)
        self.assertEqual(imports_charaudit.imports[0].unique_id, 'default')

        imports_both = corptools_import.get_imports_with_perms(user_both)
        self.assertEqual(len(imports_both.imports), 2)

    def test_get_imports_with_perms_ignore(self):
        app_import = import_apps()['allianceauth.authentication']

        AppSettings.objects.filter(
            Q(app_name='marketmanager_corporation') | Q(app_name__startswith='corptools')
        ).update(ignored=True)

        imports = app_import.get_imports_with_perms(self.user)
        self.assertEqual(len(imports.imports), 1)

        user_corp = UserMainFactory(permissions=_corp_perms)
        user_charaudit = UserMainFactory(permissions=['corptools.view_characteraudit'])
        user_both = UserMainFactory(permissions=['corptools.view_characteraudit', *_corp_perms])

        corptools_import = import_apps()['corptools']

        imports_corp = corptools_import.get_imports_with_perms(user_corp)
        self.assertEqual(len(imports_corp.imports), 0)

        imports_charaudit = corptools_import.get_imports_with_perms(user_charaudit)
        self.assertEqual(len(imports_charaudit.imports), 0)

        imports_both = corptools_import.get_imports_with_perms(user_both)
        self.assertEqual(len(imports_both.imports), 0)

        user_marketmanager = UserMainFactory(permissions=['marketmanager.basic_market_browser'])
        marketmanager_import = import_apps()['marketmanager']

        imports_marketmanager = marketmanager_import.get_imports_with_perms(user_marketmanager)
        self.assertEqual(len(imports_marketmanager.imports), 1)
        self.assertEqual(imports_marketmanager.imports[0].unique_id, 'character')

    def test_has_any_perms(self):
        app_import = import_apps()['allianceauth.authentication']
        self.assertTrue(app_import.has_any_perms(self.user))

    def test_has_any_perms_ignore(self):
        app_import = import_apps()['allianceauth.authentication']

        AppSettings.objects.filter(
            Q(app_name='marketmanager_corporation') | Q(app_name__startswith='corptools')
        ).update(ignored=True)

        self.assertTrue(app_import.has_any_perms(self.user))

        user_marketmanager = UserMainFactory(permissions=['marketmanager.basic_market_browser'])
        marketmanager_import = import_apps()['marketmanager']
        self.assertTrue(marketmanager_import.has_any_perms(user_marketmanager))

        user_both = UserMainFactory(permissions=['corptools.view_characteraudit', *_corp_perms])
        corptools_import = import_apps()['corptools']
        self.assertFalse(corptools_import.has_any_perms(user_both))

    def test_get_ok(self):
        app_import = import_apps()['allianceauth.authentication']
        login_import = app_import.get('default')
        self.assertEqual(login_import.unique_id, 'default')

    def test_get_not_found(self):
        app_import = import_apps()['allianceauth.authentication']
        with self.assertRaises(KeyError):
            app_import.get('not_found')

    def test_validate_import(self):
        app_import = import_apps()['allianceauth.authentication']
        app_import.validate_import()

        app_import.app_label = 1
        with self.assertRaises(AssertionError):
            app_import.validate_import()
        app_import.app_label = 'allianceauth.authentication'

        tmp = app_import.imports
        app_import.imports = 1
        with self.assertRaises(AssertionError):
            app_import.validate_import()
        app_import.imports = []
        with self.assertRaises(AssertionError):
            app_import.validate_import()
        app_import.imports = tmp

        app_import.validate_import()
        app_import.imports[0].unique_id = 1
        with self.assertRaises(AssertionError):
            app_import.validate_import()
        app_import.imports[0].unique_id = 'default'

        app_import.imports[0].unique_id = 'invalid id'
        with self.assertRaises(AssertionError):
            app_import.validate_import()
        app_import.imports[0].unique_id = 'default'

        app_import.imports[0].add_character = 1
        with self.assertRaises(AssertionError):
            app_import.validate_import()
        app_import.imports[0].add_character = lambda request, token: None

        app_import.imports[0].scopes = 1
        with self.assertRaises(AssertionError):
            app_import.validate_import()
        app_import.imports[0].scopes = ['publicData']

        app_import.imports[0].check_permissions = 1
        with self.assertRaises(AssertionError):
            app_import.validate_import()
        app_import.imports[0].check_permissions = lambda user: True

        tmp = app_import.imports[0].is_character_added
        app_import.imports[0].is_character_added = 1
        with self.assertRaises(AssertionError):
            app_import.validate_import()
        app_import.imports[0].is_character_added = tmp

        tmp = app_import.imports[0].is_character_added_annotation
        app_import.imports[0].is_character_added_annotation = 1
        with self.assertRaises(AssertionError):
            app_import.validate_import()
        app_import.imports[0].is_character_added_annotation = tmp

        tmp = app_import.imports[0].get_users_with_perms
        app_import.imports[0].get_users_with_perms = 1
        with self.assertRaises(AssertionError):
            app_import.validate_import()
        app_import.imports[0].get_users_with_perms = tmp

        app_import.imports.append(app_import.imports[0])
        with self.assertRaises(AssertionError):
            app_import.validate_import()
        app_import.imports = [app_import.imports[0]]

        app_import.imports[0].app_label = 'different_app'
        with self.assertRaises(AssertionError):
            app_import.validate_import()
        app_import.imports[0].app_label = 'allianceauth.authentication'

        app_import.app_label = 'different_app'
        with self.assertRaises(AssertionError):
            app_import.validate_import()
        app_import.app_label = 'allianceauth.authentication'



from unittest.mock import patch

from django.test import TestCase, RequestFactory
from django.contrib.messages.storage.fallback import FallbackStorage

from app_utils.testdata_factories import UserMainFactory, EveCorporationInfoFactory

from charlink.imports.moonmining import _add_character, _is_character_added
from charlink.app_imports import import_apps


class TestAddCharacter(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=[
            "moonmining.add_refinery_owner",
            "moonmining.basic_access"
        ])
        cls.character = cls.user.profile.main_character

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()

    @patch('moonmining.tasks.update_owner.delay')
    def test_ok(self, mock_update_owner):
        mock_update_owner.return_value = None

        token = self.user.token_set.first()

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages

        _add_character(request, token)

        mock_update_owner.assert_called_once()
        self.assertTrue(_is_character_added(self.character))

    @patch('allianceauth.eveonline.managers.EveCorporationManager.create_corporation', wraps=lambda corp_id: EveCorporationInfoFactory(corporation_id=corp_id))
    @patch('moonmining.tasks.update_owner.delay')
    def test_missing_corporation(self, mock_update_owner, mock_create_corporation):
        mock_update_owner.return_value = None

        self.character.corporation.delete()

        token = self.user.token_set.first()

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages

        _add_character(request, token)

        mock_update_owner.assert_called_once()
        mock_create_corporation.assert_called_once()
        self.assertTrue(_is_character_added(self.character))

    @patch('moonmining.tasks.update_owner.delay')
    @patch('charlink.imports.moonmining.MOONMINING_ADMIN_NOTIFICATIONS_ENABLED', False)
    @patch('charlink.imports.moonmining.notify_admins')
    def test_no_admin_notification(self, mock_notify_admins, mock_update_owner):
        mock_update_owner.return_value = None
        mock_notify_admins.return_value = None

        token = self.user.token_set.first()

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages

        _add_character(request, token)

        self.assertFalse(mock_notify_admins.called)
        mock_update_owner.assert_called_once()
        self.assertTrue(_is_character_added(self.character))


class TestIsCharacterAdded(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=[
            "moonmining.add_refinery_owner",
            "moonmining.basic_access"
        ])
        cls.character = cls.user.profile.main_character
        cls.token = cls.user.token_set.first()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()

    @patch('moonmining.tasks.update_owner.delay')
    def test_ok(self, mock_update_owner):
        mock_update_owner.return_value = None

        request = self.factory.get('/charlink/login/')
        setattr(request, 'session', 'session')
        messages = FallbackStorage(request)
        request._messages = messages

        self.assertFalse(_is_character_added(self.character))
        _add_character(request, self.token)
        self.assertTrue(_is_character_added(self.character))


class TestCheckPermissions(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.no_perm_user = UserMainFactory()
        cls.perm_user = UserMainFactory(permissions=["moonmining.add_refinery_owner", "moonmining.basic_access"])

    def test_ok(self):
        login_import = import_apps()['moonmining'].get('default')

        self.assertTrue(login_import.check_permissions(self.perm_user))
        self.assertFalse(login_import.check_permissions(self.no_perm_user))


class TestGetUsersWithPerms(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.no_perm_user = UserMainFactory()
        cls.perm_user = UserMainFactory(permissions=["moonmining.add_refinery_owner", "moonmining.basic_access"])

    def test_ok(self):
        login_import = import_apps()['moonmining'].get('default')

        users = login_import.get_users_with_perms()
        self.assertEqual(users.count(), 1)
        self.assertEqual(users.first(), self.perm_user)
